(*******************************************************************************
  Copyright 2014-2015 MIT Lincoln Laboratory
  Project:     SPAR
  Author:      Alley Stoughton
  Description: Part of Private Count Retrieval (PCR) Protocol Proof
*******************************************************************************)

(* Client.ec *)

(* Proof of Security Against Client *)

prover "Alt-Ergo". timeout 10.  (* set prover and timeout *)

(******************** PCR Protocol and Related Definitions ********************)

require import Protocol.

(*************************** Budgeted Random Oracle ***************************)

(* hashing budget *)

op budget : int.

(* the following axiom implies that hash collisions aren't *forced*,
   if no more than budget distinct element/secret pairs are hashed *)

axiom Budget : 0 <= budget <= 2 ^ tagLen.

clone RO.BudgetedRandomOracle as BRO with
  op budget <- budget
proof *.
(* beginning of realization *)
realize OutputDistrMinusWeight. apply TagDistrMinusWeight. qed.

realize Budget. apply Budget. qed.
(* end of realization *)

(*************************** Elements Counts Maps ****************************)

(* counts for certain elements; counts are always >= 1 *)

type elems_counts = (elem, int)map.

(* empty elements counts map *)

op emptyEC : elems_counts = FMap.empty.

(* when looking up a count, the default result is 0 *)

op getCount (elemsCounts : elems_counts) (elem : elem) : int =
  if mem elem (dom elemsCounts) then oget elemsCounts.[elem] else 0.

(* when incrementing a count, if the element isn't already in the
   map's domain, we start at 1 *)

op incrCount (elemsCounts : elems_counts) (elem : elem) : elems_counts =
  if mem elem (dom elemsCounts)
  then elemsCounts.[elem <- oget(elemsCounts.[elem]) + 1]
  else elemsCounts.[elem <- 1].

(* lemmas about elements counts maps *)

lemma ECGetNotInDom (elemsCounts : elems_counts) (elem : elem) :
  ! mem elem (dom elemsCounts) =>
  getCount elemsCounts elem = 0
by smt.

lemma ECGetInDom (elemsCounts : elems_counts) (elem : elem) :
  mem elem (dom elemsCounts) =>
  getCount elemsCounts elem = oget elemsCounts.[elem]
by smt.

lemma ECGetEmpty (elem : elem) :
  getCount emptyEC elem = 0
by smt.

lemma ECIncrGetEqual (elemsCounts : elems_counts) (elem : elem) :
  getCount (incrCount elemsCounts elem) elem =
  getCount elemsCounts elem + 1.
proof.
rewrite /getCount /incrCount; smt.
qed.

lemma ECIncrGetNotEqual (elemsCounts : elems_counts) (elem elem' : elem) :
  elem' <> elem =>
  getCount (incrCount elemsCounts elem) elem' =
  getCount elemsCounts elem'.
proof.
rewrite /getCount /incrCount; smt.
qed.

lemma ECIncrDom (elemCounts) (elem : elem) :
  dom(incrCount elemCounts elem) = add elem (dom elemCounts)
by smt.

lemma ECInvar (elemsCounts : elems_counts) (elem : elem) (xs : elem list) :
  mem elem xs =>
  (forall (elem' : elem),
   mem elem' (dom elemsCounts) =>
   mem elem' xs /\ oget elemsCounts.[elem'] >= 1) =>
  (forall (elem' : elem),
   mem elem' (dom(incrCount elemsCounts elem)) =>
   mem elem' xs /\ oget (incrCount elemsCounts elem).[elem'] >= 1).
proof.
intros H1 H2 elem' H3.
cut H4 :
  dom(incrCount elemsCounts elem) = add elem (dom elemsCounts) by smt.
cut H5 :
  elem' = elem \/
  (elem' <> elem /\ mem elem' (dom elemsCounts)) by smt.
split; [smt | rewrite /incrCount; smt].
qed.

(********************************* Adversary **********************************)

(* Adversary is parameterized by a budgeted random oracle

   its pre procedure takes in the Server's secret and returns a
   database and list of queries, possibly using O.bhash (budgeted
   hashing) in the process

   its post procedure passes judgment on the view it's given, possibly
   using O.hash (non-budgeted hashing) in the process *)

module type ADV(O : BRO.BOR) = {
  proc * pre(sec : sec) : db * qrys {O.bhash}
  proc post(view : client_view) : bool {O.hash}
}.

(*************************** Real and Ideal Games *****************************)

(* the "real" game

   Server and Client do hashing using random oracle

   Adversary is given budgeted random oracle derived from random oracle

   Adversary's pre procedure does budgeted hashing; if it doesn't leave enough
   budget for Server/Client, it loses game

   Adversary's post procedure does unbudgeted hashing

   hash collisions may result in inflated counts for Client's queries *)

module GReal(Adv : ADV) : GAME = {
  module Or  = RO.Or        (* random oracle *)
  module BOr = BRO.BOr(Or)  (* budgeted random oracle built from Or *)

  module A = Adv(BOr)

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b, suff : bool;
    var sec : sec;
    var hdb : hdb;
    var view : client_view;
    Protocol.initViews();
    BOr.init();  (* also initializes Or *)
    sec = Protocol.serverGenSec();
    (db, qrys) = A.pre(sec);
    suff = BOr.suff(length db  + length qrys);
    if (suff) {  (* did Adversary live within budget? *)
      hdb = Protocol.serverHashDB(db);
      Protocol.tpAcceptHDB(hdb);
      Protocol.clientLoop(qrys);
      view = Protocol.clientGetView();
    }
    else {
      view = [];  (* Adversary loses *)
    }
    b = A.post(view);
    return b;
  }
}.

(* the "ideal" game

   Adversary is given budgeted random oracle derived from random oracle

   Adversary's pre procedure does budgeted hashing; if it doesn't leave enough
   budget for Server/Client, it loses game

   Adversary's post procedure does unbudgeted hashing

   Client is given map containing (correct) counts for its queries,
   which it consults instead of interacting with TP; Client must
   still do hashing of its queries (using random oracle)

   Server doesn't do any hashing, and TP is removed
*)

module GIdeal(Adv : ADV) : GAME = {
  module Or  = RO.Or        (* random oracle *)
  module BOr = BRO.BOr(Or)  (* budgeted random oracle built from Or *)

  (* View *)

  var clientView : client_view  (* Client's view *)

  proc clientInitView() : unit = {  (* must be called first, only once *)
    clientView = [];
  }

  proc clientGetView() : client_view = {
    return clientView;
  }

  (* Server *)

  var servSec : sec  (* the Server's secret *)

  proc serverGenSec() : sec = {
    servSec = $secDistr;
    return servSec;
  }

  proc serverGetSec() : sec = { (* called by clientLoop *)
    return servSec;
  }

  (* Query Counting *)

  proc computeAnswers(qrys : qrys, db : db) : elems_counts = {
    var qrysCounts : elems_counts;
    var i : int;
    var elem : elem;
    qrysCounts = emptyEC;
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      if (mem elem qrys) {
        qrysCounts = incrCount qrysCounts elem;
      }
      i = i + 1;
    }
    return qrysCounts;
  }

  (* Client *)

  proc clientLoop(qrys : qrys, qrysCounts : elems_counts) : unit = {
    var i, n : int;
    var tag : tag;
    var qry : elem;
    var sec : sec;
    clientView = clientView ++ [ClientViewQrys qrys];
    sec = serverGetSec();
    clientView = clientView ++ [ClientViewGotSec sec];
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = Or.hash((qry, sec));
      n = getCount qrysCounts qry;
      clientView = clientView ++ [ClientViewQryCount(qry, tag, n)];
      i = i + 1;
    }
  }

  (* Main Processing *)

  module A = Adv(BOr)

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b, suff : bool;
    var sec : sec;
    var qrysCounts : elems_counts;
    var view : client_view;
    clientInitView();
    BOr.init();  (* also initializes Or *)
    sec = serverGenSec();
    (db, qrys) = A.pre(sec);
    suff = BOr.suff(length db  + length qrys);
    if (suff) {  (* did Adversary live within budget? *)
      qrysCounts = computeAnswers(qrys, db);
      clientLoop(qrys, qrysCounts);
      view = clientGetView();
    }
    else {
      view = [];  (* Adversary loses *)
    }
    b = A.post(view);
    return b;
  }
}.

(* see end-of-file for top-level theorem *)

(********************************* Proof Body *********************************)

(************************* Theories Supporting Proof **************************)

(* disjoint unions *)

require import Union.

(* injective maps *)

require import Inj.

(* numbers of occurrences of elements in lists *)

require import NumOccs.

(******************************** Proof Section *******************************)

(* the rest of the proof is within a section, in which the Adversary,
   Adv, is declared locally *)

section.

declare module Adv : ADV{BRO.BOr(RO.Or), GReal, GIdeal}.

(* H1 is like GReal, except

   - its Adversary is not a parameter;

   - it takes a budgeted random oracle as its parameter, and
     serverHashDB and clientQueryLoop use budgeted hashing

   - inlining and dead code elimination have been done *)

local module H1(O : BRO.BOR) = {
  var view : client_view
  var sec : sec
  var hdb : hdb

  proc serverHashDB(db : db) : hdb = {
    var i : int;
    var elem : elem;
    var tag : tag;
    var hdb : tag list;
    db = Shuffle.shuffle(db);
    hdb = [];
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      tag = O.bhash((elem, sec));  (* budgeted *)
      hdb = hdb ++ [tag];
      i = i + 1;
    }
    return hdb;
  }

  proc clientLoop(qrys : qrys) : unit = {
    var i, j, n : int;
    var tag : tag;
    var qry : elem;
    view = view ++ [ClientViewQrys qrys; ClientViewGotSec sec];
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = O.bhash((qry, sec));  (* budgeted *)
      j = 0;
      n = 0;
      while (j < length hdb) {
        if (oget(nth hdb j) = tag) {
          n = n + 1;
        }
        j = j + 1;
      }
      view = view ++ [ClientViewQryCount(qry, tag, n)];
      i = i + 1;
    }
  }

  module A = Adv(O)

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b, suff : bool;
    view = [];
    O.init();
    sec = $secDistr;
    (db, qrys) = A.pre(sec);
    suff = O.suff(length db  + length qrys);
    if (suff) {
      hdb = serverHashDB(db);
      clientLoop(qrys);
    }
    b = A.post(view);
    return b;
  }
}.

(* G1 is identical to H1 except it uses oracle BRO.Or, instead of
   being parameterized by an oracle *)

local module G1 = H1(BRO.BOr(RO.Or)).

local lemma Protocol_G1_serverHashDB :
  equiv
  [Protocol.serverHashDB ~ G1.serverHashDB :
   ={db, RO.Or.mp} /\ Protocol.serverSec{1} = G1.sec{2} ==>
   ={res, RO.Or.mp}].
proof.
proc.
sim (: ={RO.Or.mp}) / true : (={RO.Or.mp, hdb}).
apply BRO.Or_BOr_Or_hash_bhash.
qed.

local lemma Protocol_G1_clientLoop :
  equiv
  [Protocol.clientLoop ~ G1.clientLoop :
   ={qrys, RO.Or.mp} /\ Protocol.serverSec{1} = G1.sec{2} /\
   Protocol.clientView{1} = G1.view{2} /\
   Protocol.tpHDB{1} = G1.hdb{2} ==>
   ={RO.Or.mp} /\
   Protocol.clientView{1} = G1.view{2}].
proof.
proc; inline Protocol.serverGetSec Protocol.tpCountTag.
(* initialization of view, plus assignment to sec in (1) *)
seq 3 1 :
  (={qrys, RO.Or.mp} /\ sec{1} = G1.sec{2} /\
   Protocol.clientView{1} = G1.view{2} /\
   Protocol.tpHDB{1} = G1.hdb{2}).
auto; progress; rewrite - appA; trivial.
(* outer while loop *)
while
  (={qrys, i, RO.Or.mp} /\ sec{1} = G1.sec{2} /\
   Protocol.clientView{1} = G1.view{2} /\
   Protocol.tpHDB{1} = G1.hdb{2}).
wp.
(* inner while loop *)
while
  (i0{1} = j{2} /\ n0{1} = n{2} /\ tag0{1} = tag{2} /\
   Protocol.tpHDB{1} = G1.hdb{2}).
auto.
(* beginning of outer while loop *)
wp.
call BRO.Or_BOr_Or_hash_bhash.
auto.
(* before outer while loop *)
auto.
qed.

local lemma GReal_G1_main :
  equiv[GReal(Adv).main ~ G1.main : true ==> ={res}].
proof.
proc.
(* call to A.post *)
call (_ : ={view, RO.Or.mp, glob Adv} ==> ={res, RO.Or.mp}).
proc (={RO.Or.mp}) => //.
sim.
(* rest *)
inline Protocol.initViews Protocol.serverGenSec Protocol.serverGetSec
       Protocol.tpAcceptHDB.
(* initialization, call to pre, calculation of suff *)
swap {1} 3 6; swap {2} 1 4.
seq 8 4 :
  (={db, qrys, suff, glob BRO.BOr(RO.Or), glob Adv} /\
   Protocol.serverSec{1} = G1.sec{2} /\ sec{1} = G1.sec{2}).
sim.
sp.
(* conditional *)
if => //.
(* true branch *)
(* serverHashDB *)
seq 4 1 :
  (={db, qrys, RO.Or.mp, glob Adv} /\ Protocol.serverSec{1} = G1.sec{2} /\
   Protocol.clientView{1} = G1.view{2} /\ Protocol.tpHDB{1} = G1.hdb{2}).
wp.
call Protocol_G1_serverHashDB.
auto.
(* deal with clientGetView *)
inline Protocol.clientGetView; wp.
(* clientQueryLoop *)
call Protocol_G1_clientLoop.
auto.
(* false branch *)
auto.
qed.

local lemma GReal_G1 &m :
  Pr[GReal(Adv).main() @ &m : res] = Pr[G1.main() @ &m : res].
proof.
byequiv GReal_G1_main; trivial.
qed.

(* we locally clone the abstract theory BRO.Switching as the theory
   Switching, fully realizing it (there was nothing to realize) *)

local clone BRO.Switching as Switching
proof *.
(* nothing to realize *)

(* G2 is the same as G1, except it uses Switching.OrInj not BRO.BOr(RO.Or) *)

local module G2 = H1(Switching.BOrInj).

(* we apply the Switching lemma to G1/G2

   our concrete switching adversary H1' is the same as H1, except that
   its main procedure doesn't (can't!) initialize the oracle, and hdb
   is always initialized *)

local module H1'(O : BRO.BOR) : Switching.SWADV(O) = {
  var view : client_view
  var sec : sec
  var hdb : hdb

  proc serverHashDB(db : db) : hdb = {
    var i : int;
    var elem : elem;
    var tag : tag;
    var hdb : tag list;
    db = Shuffle.shuffle(db);
    hdb = [];
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      tag = O.bhash((elem, sec));
      hdb = hdb ++ [tag];
      i = i + 1;
    }
    return hdb;
  }

  proc clientLoop(qrys : qrys) : unit = {
    var i, j, n : int;
    var tag : tag;
    var qry : elem;
    view = view ++ [ClientViewQrys qrys; ClientViewGotSec sec];
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = O.bhash((qry, sec));
      j = 0;
      n = 0;
      while (j < length hdb) {
        if (oget(nth hdb j) = tag) {
          n = n + 1;
        }
        j = j + 1;
      }
      view = view ++ [ClientViewQryCount(qry, tag, n)];
      i = i + 1;
    }
  }

  module A = Adv(O)

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b, suff : bool;
    view = [];
    (* no oracle initialization *)
    sec = $secDistr;
    hdb = [];  (* new *)
    (db, qrys) = A.pre(sec);
    suff = O.suff(length db  + length qrys);
    if (suff) {
      hdb = serverHashDB(db);
      clientLoop(qrys);
    }
    b = A.post(view);
    return b;
  }
}.

local lemma H1'_Lossless_serverHashDB :
  forall (O <: BRO.BOR),
  islossless O.bhash =>
  islossless H1'(O).serverHashDB.
proof.
intros O LossBHash.
proc.
while (true) (length db - i).
intros z.
wp.
call (_ : true ==> true); first apply LossBHash.
auto; smt.
wp.
call (_ : true ==> true); first apply Lossless_shuffle.
auto; smt.
qed.

local lemma H1'_Lossless_clientLoop :
  forall (O <: BRO.BOR{H1'}),
  islossless O.bhash =>
  islossless H1'(O).clientLoop.
proof.
intros O LossBHash.
proc.
while (true) (length qrys - i).
intros z.
wp.
while (true) (length H1'.hdb - j).
intros z0; auto; smt.
wp.
call (_ : true ==> true); first apply LossBHash.
wp.
skip; smt.
wp.
skip; smt.
qed.

local lemma H1'_Lossless_main :
  (forall (O <: BRO.BOR{Adv}),
   islossless O.bhash => islossless Adv(O).pre) =>
  (forall (O <: BRO.BOR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  forall (O <: BRO.BOR{H1'}),
  islossless O.bhash => islossless O.hash => islossless O.suff =>
  islossless H1'(O).main.
proof.
intros LossPre LossPost O LossBHash LossHash LossSuff.
proc.
(* call to pre and suff *)
seq 5 : true.
(* true postcondition *)
call (_ : true ==> true).
proc (true); trivial.
call (_ : true ==> true).
proc (true); trivial.
wp; rnd.
auto.
(* losslessness *)
call (_ : true ==> true).
apply LossSuff.
call (_ : true ==> true).
apply (LossPre O); first apply LossBHash.
sp; wp; rnd.
skip; smt.
(* rest of main, true postcondition *)
if.
(* then branch *)
call (_ : true ==> true).
apply (LossPost O); first apply LossHash.
call (_ : true ==> true).
apply (H1'_Lossless_clientLoop O); first apply LossBHash.
call (_ : true ==> true).
apply (H1'_Lossless_serverHashDB O); first apply LossBHash.
auto.
(* else branch *)
call (_ : true ==> true).
apply (LossPost O); first apply LossHash.
auto.
(* first five lines can't result in false *)
hoare.
call (_ : true ==> true).
proc (true); trivial.
call (_ : true ==> true).
proc (true); trivial.
auto.
(* probability agreement *)
trivial.
qed.

(* now we apply the Switching Lemma to H1' *)

local lemma Switching_H1' &m :
  (forall (O <: BRO.BOR{Adv}),
   islossless O.bhash => islossless Adv(O).pre) =>
  (forall (O <: BRO.BOR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  `|Pr[Switching.GSwitching(H1', BRO.BOr(RO.Or)).main() @ &m : res] -
    Pr[Switching.GSwitching(H1', Switching.BOrInj).main() @ &m : res]| <=
  Switching.collBound.
proof.
intros LossPre LossPost.
apply (Switching.Switching H1' &m).
apply H1'_Lossless_main; [apply LossPre | apply LossPost].
qed.

local lemma G1_GSwitching_H1'_Or_main &m :
  equiv
  [G1.main ~ Switching.GSwitching(H1', BRO.BOr(RO.Or)).main :
   true ==> ={res}].
proof.
proc.
inline Switching.GSwitching(H1', BRO.BOr(RO.Or)).SA.main.
swap {1} 1; wp; sim.
qed.

local lemma G1_GSwitching_H1'_Or &m :
  Pr[G1.main() @ &m : res] =
  Pr[Switching.GSwitching(H1', BRO.BOr(RO.Or)).main() @ &m : res].
proof.
byequiv (G1_GSwitching_H1'_Or_main &m) => //.
qed.

local lemma GSwitching_H1'_OrInj_G2_main &m :
  equiv
  [Switching.GSwitching(H1', Switching.BOrInj).main ~ G2.main:
   true ==> ={res}].
proof.
proc.
inline Switching.GSwitching(H1', Switching.BOrInj).SA.main.
swap {1} 1; sim.
qed.

local lemma GSwitching_H1'_OrInj_G2 &m :
  Pr[Switching.GSwitching(H1', Switching.BOrInj).main() @ &m : res] =
  Pr[G2.main() @ &m : res].
proof.
byequiv (GSwitching_H1'_OrInj_G2_main &m) => //.
qed.

local lemma G1_G2 &m :
  (forall (O <: BRO.BOR{Adv}),
   islossless O.bhash => islossless Adv(O).pre) =>
  (forall (O <: BRO.BOR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  `|Pr[G1.main() @ &m : res] - Pr[G2.main() @ &m : res]| <=
  Switching.collBound.
proof.
intros LossPre LossPost.
rewrite (G1_GSwitching_H1'_Or &m).
rewrite - (GSwitching_H1'_OrInj_G2 &m).
apply (Switching_H1' &m); [apply LossPre | apply LossPost].
qed.

local lemma GReal_G2 &m :
  (forall (O <: BRO.BOR{Adv}),
   islossless O.bhash => islossless Adv(O).pre) =>
  (forall (O <: BRO.BOR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  `|Pr[GReal(Adv).main() @ &m : res] - Pr[G2.main() @ &m : res]| <=
   Switching.collBound.
proof.
intros LossPre LossPost.
rewrite (GReal_G1 &m).
apply (G1_G2 &m); [apply LossPre | apply LossPost].
qed.

(* In H2, the count for each query is exactly the number of times the
   query occurs in the database

   serverHashDB is renamed to serverHashAndCountDB; it still hashes
   the database, but no longer constructs the hashed database (and
   thus the global variable hdb is deleted) -- instead it produces a
   map containing the counts for all of the database's elements

   clientQueryLoop now takes the resulting element counts map as an
   argument, and consults this map instead of counting the number of
   occurrences of a query's tag in the hashed database; but it
   still hashes each query *)

local module H2(O : BRO.BOR) = {
  var view : client_view
  var sec : sec

  proc serverHashAndCountDB(db : db) : elems_counts = {
    var i : int;
    var elem : elem;
    var tag : tag;
    var elemsCounts : elems_counts;
    elemsCounts = emptyEC;
    db = Shuffle.shuffle(db);
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      tag = O.bhash((elem, sec));
      elemsCounts = incrCount elemsCounts elem;
      i = i + 1;
    }
    return elemsCounts;
  }

  proc clientLoop(qrys : qrys, elemsCounts : elems_counts) : unit = {
    var i, n : int;
    var tag : tag;
    var qry : elem;
    view = view ++ [ClientViewQrys qrys; ClientViewGotSec sec];
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = O.bhash((qry, sec));
      n = getCount elemsCounts qry;
      view = view ++ [ClientViewQryCount(qry, tag, n)];
      i = i + 1;
    }
  }

  module A = Adv(O)

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b, suff : bool;
    var elemsCounts : elems_counts;
    view = [];
    O.init();
    sec = $secDistr;
    (db, qrys) = A.pre(sec);
    suff = O.suff(length db  + length qrys);
    if (suff) {
      elemsCounts = serverHashAndCountDB(db);
      clientLoop(qrys, elemsCounts);
    }
    b = A.post(view);
    return b;
  }
}.

(* G3 is like H2 except it uses oracle Switching.BOrInj, instead of being
   parameterized by an oracle *)

local module G3 = H2(Switching.BOrInj).

(* to prove G2 equivalent to G3, we need an invariant relating
   the secret, hashed database (in G2), oracle's map, and elements
   counts map (in G3)

   see NumOccs.ec for the definition of numOccs

   below, we'll also use the predicate Inj of Inj.ec for injective
   (collision free) maps *)

pred Counting
     (sec : sec, hdb : hdb, mp : (elem * sec, tag)map,
      elemsCounts : elems_counts) =
  (forall (tag : tag),
   mem tag hdb => mem tag (rng mp)) /\
  (forall (elem : elem),
   !(mem elem (dom elemsCounts)) =>
   !(mem (elem, sec) (dom mp)) \/
   numOccs (oget mp.[(elem, sec)]) hdb = 0) /\
  (forall (elem : elem),
   mem elem (dom elemsCounts) =>
   mem (elem, sec) (dom mp) /\
   numOccs (oget mp.[(elem, sec)]) hdb =
   oget elemsCounts.[elem]).

local lemma Counting_Intro
  (sec : sec, hdb : hdb, mp : (elem * sec, tag)map,
   elemsCounts : elems_counts) :
  (forall (tag : tag),
   mem tag hdb => mem tag (rng mp)) /\
  (forall (elem : elem),
   !(mem elem (dom elemsCounts)) =>
   !(mem (elem, sec) (dom mp)) \/
   numOccs (oget mp.[(elem, sec)]) hdb = 0) /\
  (forall (elem : elem),
   mem elem (dom elemsCounts) =>
   mem (elem, sec) (dom mp) /\
   numOccs (oget mp.[(elem, sec)]) hdb =
   oget elemsCounts.[elem]) =>
  Counting sec hdb mp elemsCounts
by trivial.

local lemma Counting_Elim
  (sec : sec, hdb : hdb, mp : (elem * sec, tag)map,
   elemsCounts : elems_counts) :
  Counting sec hdb mp elemsCounts =>
  (forall (tag : tag),
   mem tag hdb => mem tag (rng mp)) /\
  (forall (elem : elem),
   !(mem elem (dom elemsCounts)) =>
   !(mem (elem, sec) (dom mp)) \/
   numOccs (oget mp.[(elem, sec)]) hdb = 0) /\
  (forall (elem : elem),
   mem elem (dom elemsCounts) =>
   mem (elem, sec) (dom mp) /\
   numOccs (oget mp.[(elem, sec)]) hdb =
   oget elemsCounts.[elem]).
proof.
intros H => //.
qed.

local lemma Counting_InECDom_Implies_InDomOracle
  (sec : sec, hdb : hdb, mp : (elem * sec, tag)map,
   elemsCounts : elems_counts, elem : elem) :
  Counting sec hdb mp elemsCounts =>
  mem elem (dom elemsCounts) =>
  mem (elem, sec) (dom mp)
by smt.

local lemma Counting_Empty (sec : sec, mp : (elem * sec, tag)map) :
  Counting sec [] mp emptyEC
by smt.

local lemma Counting_AddOracleMap
  (sec : sec, hdb : hdb, mp : (elem * sec, tag)map, elemsCounts : elems_counts,
   elem : elem, tag : tag) :
  Counting sec hdb mp elemsCounts =>
  !(mem (elem, sec) (dom mp)) =>
  !(mem tag (rng mp)) =>
  Counting sec hdb (mp.[(elem, sec) <- tag]) elemsCounts.
proof.
intros H1 H2 H3.
progress; smt.
qed.

local lemma Counting_AddHDB_DiffElem_NotInEC
  (sec : sec, hdb : hdb, mp : (elem * sec, tag)map,
   elemsCounts : elems_counts, elem elem' : elem) :
  Inj mp (dom mp) (rng mp) =>
  Counting sec hdb mp elemsCounts =>
  elem' <> elem =>
  mem (elem, sec) (dom mp) =>
  !(mem elem' (dom elemsCounts)) =>
  !(mem (elem', sec) (dom mp)) \/
  numOccs (oget mp.[(elem', sec)]) (hdb ++ [oget mp.[(elem, sec)]]) = 0
by smt.

local lemma Counting_AddHDB_DiffElem_InEC
  (sec : sec, hdb : hdb, mp : (elem * sec, tag)map,
   elemsCounts : elems_counts, elem elem' : elem) :
  Inj mp (dom mp) (rng mp) =>
  Counting sec hdb mp elemsCounts =>
  elem' <> elem =>
  mem (elem, sec) (dom mp) =>
  mem elem' (dom elemsCounts) =>
  mem (elem', sec) (dom mp) /\
  numOccs (oget mp.[(elem', sec)]) (hdb ++ [oget mp.[(elem, sec)]]) =
  oget elemsCounts.[elem']
by smt.

local lemma Counting_Add
  (sec : sec, hdb : hdb, mp : (elem * sec, tag)map,
   elemsCounts : elems_counts, elem : elem) :
  Inj mp (dom mp) (rng mp) =>
  Counting sec hdb mp elemsCounts =>
  mem (elem, sec) (dom mp) =>
  Counting
  sec (hdb ++ [oget mp.[(elem, sec)]]) mp
  (incrCount elemsCounts elem).
proof.
intros Inj [Cntg1 [Cntg2 Cntg3]] DomMP.
rewrite /incrCount.
split; first smt.
split; first smt.
progress; smt.
qed.

local lemma OrInj_Counting_bhash
  (sec : sec, hdb : hdb, elemsCounts : elems_counts, inp' : elem * sec,
   ctr' : int) :
  equiv
  [Switching.BOrInj.bhash ~ Switching.BOrInj.bhash :
   ={inp, glob Switching.BOrInj} /\ inp{1} = inp' /\
   inp{1} = (fst(inp{1}), sec) /\ Switching.BOrInj.ctr{1} = ctr' /\
   Switching.BOrInj.ctr{1} < budget /\
   dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1} /\
   ! Switching.BOrInj.over{1} /\
   Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
       Switching.BOrInj.outs{1} /\
   Counting sec hdb Switching.BOrInj.mp{1} elemsCounts ==>
   ={res, glob Switching.BOrInj} /\ mem inp' (dom Switching.BOrInj.mp{1}) /\
   oget Switching.BOrInj.mp{1}.[inp'] = res{1} /\
   Switching.BOrInj.ctr{1} <= budget /\
   dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1} /\
   ! Switching.BOrInj.over{1} /\
   Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps {1}
       Switching.BOrInj.outs{1} /\
   Switching.BOrInj.ctr{1} <= ctr' + 1 /\
   Counting sec hdb Switching.BOrInj.mp{1} elemsCounts].
proof.
proc.
if => //.
if => //.
rcondt {1} 1; first auto.
rcondt {2} 1; first auto.
wp.
seq 1 1 :
  (={inp, glob Switching.BOrInj} /\ inp{1} = inp' /\
   inp{1} = (fst(inp{1}), sec) /\
   Switching.BOrInj.ctr{1} = ctr' /\ Switching.BOrInj.ctr{1} < budget /\
   dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1} /\
   ! Switching.BOrInj.over{1} /\
   Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
       Switching.BOrInj.outs{1} /\
   Counting sec hdb Switching.BOrInj.mp{1} elemsCounts /\
   !(mem inp{1} (dom Switching.BOrInj.mp{1})) /\
   Switching.BOrInj.ctr{1} < budget /\
   ={out}).
auto.
(* test for tentative collision *)
if => //.
auto; smt.
skip; smt.
sp 1 1; if => //; auto.
skip; smt.
qed.

local lemma G2_G3_serverHashDB_serverHashAndCountDB (qrys : qrys) :
  equiv
  [G2.serverHashDB ~ G3.serverHashAndCountDB :
   ={db, glob Switching.BOrInj} /\ G2.sec{1} = G3.sec{2} /\
   dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1} /\
   Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
       Switching.BOrInj.outs{1} /\
   ! Switching.BOrInj.over{1} /\
   length db{1} + length qrys <= budget - Switching.BOrInj.ctr{1} ==>
   ={glob Switching.BOrInj} /\
   dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1} /\
   Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
       Switching.BOrInj.outs{1} /\
   ! Switching.BOrInj.over{1} /\
   Counting G2.sec{1} res{1} Switching.BOrInj.mp{1} res{2} /\
   length qrys <= budget - Switching.BOrInj.ctr{1}].
proof.
proc.
(* first handle up to while loop *)
seq 3 3 :
  (={db, i, glob Switching.BOrInj} /\ G2.sec{1} = G3.sec{2} /\ hdb{1} = [] /\
   i{1} = 0 /\ dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1} /\
   Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
       Switching.BOrInj.outs{1} /\
   ! Switching.BOrInj.over{1} /\
   length db{1} + length qrys <= budget - Switching.BOrInj.ctr{1} /\
   Counting G2.sec{1} hdb{1} Switching.BOrInj.mp{1} elemsCounts{2}).
wp; sp.
exists* db{1}; elim*; intros db_L.
call (_ : ={xs} /\ xs{1} = db_L ==> ={res} /\ length res{1} = length db_L).
apply (PreserveLength_shuffle db_L).
skip; smt.
(* now the while loop *)
while
  (={db, i, glob Switching.BOrInj} /\ G2.sec{1} = G3.sec{2} /\
   dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1} /\
   Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
       Switching.BOrInj.outs{1} /\
   ! Switching.BOrInj.over{1} /\
   Counting G2.sec{1} hdb{1} Switching.BOrInj.mp{1} elemsCounts{2} /\
   ((length db{1}) - i{1}) + length qrys <= budget - Switching.BOrInj.ctr{1} /\
   0 <= i{1} /\ i{1} <= length db{1}).
seq 1 1 :
  (={db, i, glob Switching.BOrInj} /\ G2.sec{1} = G3.sec{2} /\
   dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1} /\
   Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
       Switching.BOrInj.outs{1} /\
   ! Switching.BOrInj.over{1} /\
   Counting G2.sec{1} hdb{1} Switching.BOrInj.mp{1} elemsCounts{2} /\
   ((length db{1}) - i{1}) + length qrys <= budget - Switching.BOrInj.ctr{1} /\
   0 <= i{1} /\ i{1} <= length db{1} /\ i{1} < length db{1} /\ ={elem} /\
   elem{1} = oget(nth db{1} i{1})).
auto.
wp; simplify.
(* in preparation for call of Switching.BOrInj.bhash, we must turn five
   variables into assumptions *)
exists* G2.sec{1}; elim*; intros sec_L.
exists* hdb{1}; elim*; intros hdb_L.
exists* elemsCounts{2}; elim*; intros elemsCounts_R.
exists* elem{1}; elim*; intros elem_L.
exists* Switching.BOrInj.ctr{1}; elim*; intros ctr_L.
call
  (OrInj_Counting_bhash sec_L hdb_L elemsCounts_R
   (elem_L, sec_L) ctr_L).
skip; smt.
(* after while loop *)
skip; smt.
qed.

local lemma G2_G3_clientLoop :
  equiv
  [G2.clientLoop ~ G3.clientLoop :
   ={qrys, glob Switching.BOrInj} /\ G2.sec{1} = G3.sec{2} /\
   G2.view{1} = G3.view{2} /\
   dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1} /\
   Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
       Switching.BOrInj.outs{1} /\
   ! Switching.BOrInj.over{1} /\
   Counting G2.sec{1} G2.hdb{1} Switching.BOrInj.mp{1} elemsCounts{2} /\
   length qrys{1} <= budget - Switching.BOrInj.ctr{1} ==>
   G2.view{1} = G3.view{2} /\ ={glob Switching.BOrInj}].
proof.
proc.
(* first handle up-to outer while loop *)
seq 2 2 :
  (={qrys, glob Switching.BOrInj} /\ G2.sec{1} = G3.sec{2} /\
   G2.view{1} = G3.view{2} /\
   dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1} /\
   Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
       Switching.BOrInj.outs{1} /\
   ! Switching.BOrInj.over{1} /\
   Counting G2.sec{1} G2.hdb{1} Switching.BOrInj.mp{1} elemsCounts{2} /\
   length qrys{1} <= budget - Switching.BOrInj.ctr{1} /\
   ={i} /\ i{1} = 0).
auto.
(* now handle outer while loop *)
while
  (={qrys, i, glob Switching.BOrInj} /\ G2.sec{1} = G3.sec{2} /\
   G2.view{1} = G3.view{2} /\
   dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1} /\
   ! Switching.BOrInj.over{1} /\
   Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
       Switching.BOrInj.outs{1} /\
   Counting G2.sec{1} G2.hdb{1} Switching.BOrInj.mp{1} elemsCounts{2} /\
   0 <= i{1} /\ i{1} <= length qrys{1} /\
   length qrys{1} - i{1} <= budget - Switching.BOrInj.ctr{1}).
(* handle first line of outer while loop *)
seq 1 1 :
  (={qrys, i, glob Switching.BOrInj} /\ G2.sec{1} = G3.sec{2} /\
   G2.view{1} = G3.view{2} /\
   dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1} /\
   ! Switching.BOrInj.over{1} /\
   Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
       Switching.BOrInj.outs{1} /\
   Counting G2.sec{1} G2.hdb{1} Switching.BOrInj.mp{2} elemsCounts{2} /\
   0 <= i{1} /\ i{1} < length qrys{1} /\
   length qrys{1} - i{1} <= budget - Switching.BOrInj.ctr{1} /\ ={qry}).
auto; smt.
(* handle call to bhash *)
seq 1 1 :
  (={qrys, i, glob Switching.BOrInj} /\ G2.sec{1} = G3.sec{2} /\
   G2.view{1} = G3.view{2} /\
   dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1} /\
   Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
       Switching.BOrInj.outs{1} /\
   ! Switching.BOrInj.over{1} /\
   Counting G2.sec{1} G2.hdb{1} Switching.BOrInj.mp{1} elemsCounts{2} /\
   0 <= i{1} /\ i{1} < length qrys{1} /\
   length qrys{1} - i{1} - 1 <= budget - Switching.BOrInj.ctr{1} /\
   ={qry, tag} /\ mem (qry{1}, G2.sec{1}) (dom Switching.BOrInj.mp{1}) /\
   oget Switching.BOrInj.mp{1}.[(qry{1}, G2.sec{1})] = tag{1}).
exists* G2.sec{1}; elim*; intros sec_L.
exists* G2.hdb{1}; elim*; intros hdb_L.
exists* elemsCounts{2}; elim*; intros elemsCounts_R.
exists* qry{1}; elim*; intros qry_L.
exists* Switching.BOrInj.ctr{1}; elim*; intros ctr_L.
call
  (OrInj_Counting_bhash sec_L hdb_L elemsCounts_R
   (qry_L, sec_L) ctr_L).
skip; smt.
(* handle rest of outer while loop *)
wp.
seq 2 0 :
  (={qrys, i, glob Switching.BOrInj} /\ G2.sec{1} = G3.sec{2} /\
   G2.view{1} = G3.view{2} /\
   dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1} /\
   Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
       Switching.BOrInj.outs{1} /\
   ! Switching.BOrInj.over{1} /\
   Counting G2.sec{1} G2.hdb{1} Switching.BOrInj.mp{1} elemsCounts{2} /\
   0 <= i{1} /\ i{1} < length qrys{1} /\
   length qrys{1} - i{1} - 1 <= budget - Switching.BOrInj.ctr{1} /\
   ={qry, tag} /\ mem (qry{1}, G2.sec{1}) (dom Switching.BOrInj.mp{1}) /\
   oget Switching.BOrInj.mp{1}.[(qry{1}, G2.sec{1})] = tag{1} /\ j{1} = 0 /\
   n{1} = 0).
auto.
(* handle inner while loop of first game *)
while {1}
  (j{1} >= 0 /\ j{1} <= length G2.hdb{1} /\
   n{1} = numOccsUpTo tag{1} G2.hdb{1} j{1})
  (length G2.hdb{1} - j{1}).
(* loop termination *)
auto; smt.
(* after inner while loop *)
skip; progress; smt.
(* after outer while loop *)
skip; smt.
qed.

local lemma G2_G3_main :
  equiv[G2.main ~ G3.main : true ==> ={res}].
proof.
proc.
(* initialization including calls to pre and suff *)
seq 5 5 :
  (={glob Switching.BOrInj, glob Adv, db, qrys, suff} /\
   H1.sec{1} = H2.sec{2} /\ G2.view{1} = G3.view{2} /\
   Switching.BOrInj.ctr{1} <= budget /\
   (Switching.BOrInj.ctr{1} < budget => ! Switching.BOrInj.over{1}) /\
   (! Switching.BOrInj.over{1} =>
    dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
    Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
        Switching.BOrInj.outs{1}) /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1} /\
   suff{1} =
   (! Switching.BOrInj.over /\
    length db + length qrys <= budget - Switching.BOrInj.ctr){1}).
inline Switching.BOrInj.suff; wp.
call
  (_ :
   ={glob Switching.BOrInj, sec} /\ Switching.BOrInj.ctr{1} <= budget /\
   (Switching.BOrInj.ctr{1} < budget => ! Switching.BOrInj.over{1}) /\
   (! Switching.BOrInj.over{1} =>
    dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
    Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
        Switching.BOrInj.outs{1}) /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1} ==>
   ={res, glob Switching.BOrInj, glob Adv} /\
   Switching.BOrInj.ctr{1} <= budget /\
   (Switching.BOrInj.ctr{1} < budget => ! Switching.BOrInj.over{1}) /\
   (! Switching.BOrInj.over{1} =>
    dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
    Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
        Switching.BOrInj.outs{1}) /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1}).
proc
  (={glob Switching.BOrInj} /\ Switching.BOrInj.ctr{1} <= budget /\
   (Switching.BOrInj.ctr{1} < budget => ! Switching.BOrInj.over{1}) /\
   (! Switching.BOrInj.over{1} =>
    dom Switching.BOrInj.mp{1} = Switching.BOrInj.inps{1} /\
    Inj Switching.BOrInj.mp{1} Switching.BOrInj.inps{1}
        Switching.BOrInj.outs{1}) /\
   Switching.BOrInj.ctr{1} = card Switching.BOrInj.outs{1}) => //.
conseq Switching.BOrInj_bhash => //.
wp; rnd.
inline Switching.BOrInj.init.
auto; smt.
(* last two lines *)
if => //.
(* then part *)
call (_ : ={view, Switching.BOrInj.mp, glob Adv} ==> ={res}).
proc (={Switching.BOrInj.mp}) => //.
sim.
call G2_G3_clientLoop.
exists* qrys{1}; elim*; intros qrys_L.
call (G2_G3_serverHashDB_serverHashAndCountDB qrys_L).
skip; smt.
(* else part *)
call (_ : ={view, Switching.BOrInj.mp, glob Adv} ==> ={res}).
proc (={Switching.BOrInj.mp}) => //.
sim.
auto; smt.
qed.

local lemma G2_G3 &m :
  Pr[G2.main() @ &m : res] = Pr[G3.main() @ &m : res].
proof.
byequiv G2_G3_main; trivial.
qed.

local lemma GReal_G3 &m :
  (forall (O <: BRO.BOR{Adv}),
   islossless O.bhash => islossless Adv(O).pre) =>
  (forall (O <: BRO.BOR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  `|Pr[GReal(Adv).main() @ &m : res] - Pr[G3.main() @ &m : res]| <=
   Switching.collBound.
proof.
intros LossPre LossPost.
rewrite - (G2_G3 &m).
apply (GReal_G2 &m); [apply LossPre | apply LossPost].
qed.

(* G4 is the same as G3, except it uses BRO.BOr(RO.Or) not
   Switching.BOrInj *)

local module G4 = H2(BRO.BOr(RO.Or)).

(* we apply the Switching Lemma to G4/G3

   our concrete switching adversary H2' is the same as H2, except that
   its main procedure doesn't (can't!) initialize the oracle *)

local module H2'(O : BRO.BOR) = {
  var view : client_view
  var sec : sec

  proc serverHashAndCountDB(db : db) : elems_counts = {
    var i : int;
    var elem : elem;
    var tag : tag;
    var elemsCounts : elems_counts;
    elemsCounts = emptyEC;
    db = Shuffle.shuffle(db);
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      tag = O.bhash((elem, sec));
      elemsCounts = incrCount elemsCounts elem;
      i = i + 1;
    }
    return elemsCounts;
  }

  proc clientLoop(qrys : qrys, elemsCounts : elems_counts) : unit = {
    var i, n : int;
    var tag : tag;
    var qry : elem;
    view = view ++ [ClientViewQrys qrys; ClientViewGotSec sec];
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = O.bhash((qry, sec));
      n = getCount elemsCounts qry;
      view = view ++ [ClientViewQryCount(qry, tag, n)];
      i = i + 1;
    }
  }

  module A = Adv(O)

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b, suff : bool;
    var elemsCounts : elems_counts;
    view = [];
    (* no oracle initialization *)
    sec = $secDistr;
    (db, qrys) = A.pre(sec);
    suff = O.suff(length db  + length qrys);
    if (suff) {
      elemsCounts = serverHashAndCountDB(db);
      clientLoop(qrys, elemsCounts);
    }
    b = A.post(view);
    return b;
  }
}.

local lemma H2'_Lossless_serverHashAndCountDB :
  forall (O <: BRO.BOR),
  islossless O.bhash =>
  islossless H2'(O).serverHashAndCountDB.
proof.
intros O LossBHash.
proc.
while (true) (length db - i).
intros z.
wp.
call (_ : true ==> true); first apply LossBHash.
auto; smt.
wp; sp.
call (_ : true ==> true).
apply Lossless_shuffle.
skip; smt.
qed.

local lemma H2'_Lossless_clientLoop :
  forall (O <: BRO.BOR{H2'}),
  islossless O.bhash =>
  islossless H2'(O).clientLoop.
proof.
intros O LossBHash.
proc.
while (true) (length qrys - i).
intros z.
wp.
call (_ : true ==> true); first apply LossBHash.
auto; smt.
auto; smt.
qed.

local lemma H2'_Lossless_main :
  (forall (O <: BRO.BOR{Adv}),
   islossless O.bhash => islossless Adv(O).pre) =>
  (forall (O <: BRO.BOR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  forall (O <: BRO.BOR{H2'}),
  islossless O.bhash => islossless O.hash => islossless O.suff =>
  islossless H2'(O).main.
proof.
intros LossPre LossPost O LossBHash LossHash LossSuff.
proc.
(* choice of secret, calls to pre and suff *)
seq 4 : true.
(* true postcondition *)
call (_ : true ==> true).
proc (true) => //.
call (_ : true ==> true).
proc (true) => //.
auto.
(* losslessness *)
call (_ : true ==> true).
apply LossSuff.
call (_ : true ==> true).
apply (LossPre O); first apply LossBHash.
auto; smt.
(* rest of main *)
if.
(* then branch *)
call (_ : true ==> true).
apply (LossPost O); first apply LossHash.
call (_ : true ==> true).
apply (H2'_Lossless_clientLoop O); first apply LossBHash.
call (_ : true ==> true).
apply (H2'_Lossless_serverHashAndCountDB O); first apply LossBHash.
auto.
(* else branch *)
call (_ : true ==> true).
apply (LossPost O); first apply LossHash.
auto.
(* first four lines can't result in false *)
hoare.
call (_ : true ==> true).
proc (true) => //.
call (_ : true ==> true).
proc (true) => //.
auto.
(* probability agreement *)
trivial.
qed.

(* now we apply the Switching Lemma to H2' *)

local lemma Switching_H2' &m :
  (forall (O <: BRO.BOR{Adv}),
   islossless O.bhash => islossless Adv(O).pre) =>
  (forall (O <: BRO.BOR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  `|Pr[Switching.GSwitching(H2', BRO.BOr(RO.Or)).main() @ &m : res] -
    Pr[Switching.GSwitching(H2', Switching.BOrInj).main() @ &m : res]| <=
  Switching.collBound.
proof.
intros LossPre LossPost.
apply (Switching.Switching H2' &m).
apply H2'_Lossless_main; [apply LossPre | apply LossPost].
qed.

local lemma G3_GSwitching_H2'_OrInj_main &m :
  equiv
  [G3.main ~ Switching.GSwitching(H2', Switching.BOrInj).main :
   true ==> ={res}].
proof.
proc.
inline Switching.GSwitching(H2', Switching.BOrInj).SA.main.
swap {1} 1.
sim.
qed.

local lemma G3_GSwitching_H2'_OrInj &m :
  Pr[G3.main() @ &m : res] =
  Pr[Switching.GSwitching(H2', Switching.BOrInj).main() @ &m : res].
proof.
byequiv (G3_GSwitching_H2'_OrInj_main &m); trivial.
qed.

local lemma GSwitching_H2'_Or_G4_main &m :
  equiv
  [Switching.GSwitching(H2', BRO.BOr(RO.Or)).main ~ G4.main :
   true ==> ={res}].
proof.
proc.
inline Switching.GSwitching(H2', BRO.BOr(RO.Or)).SA.main.
swap {1} 1.
sim.
qed.

local lemma GSwitching_H2'_Or_G4 &m :
  Pr[Switching.GSwitching(H2', BRO.BOr(RO.Or)).main() @ &m : res] =
  Pr[G4.main() @ &m : res].
proof.
byequiv (GSwitching_H2'_Or_G4_main &m); trivial.
qed.

local lemma G4_G3 &m :
  (forall (O <: BRO.BOR{Adv}),
   islossless O.bhash => islossless Adv(O).pre) =>
  (forall (O <: BRO.BOR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  `|Pr[G4.main() @ &m : res] - Pr[G3.main() @ &m : res]| <=
  Switching.collBound.
proof.
intros LossPre LossPost.
rewrite - (GSwitching_H2'_Or_G4 &m).
rewrite (G3_GSwitching_H2'_OrInj &m).
apply (Switching_H2' &m); [apply LossPre | apply LossPost].
qed.

(* now we use the Triangular Inequality to show the relationship
   between GReal(Adv) and G4: *)

local lemma GReal_G4 &m :
  (forall (O <: BRO.BOR{Adv}),
   islossless O.bhash => islossless Adv(O).pre) =>
  (forall (O <: BRO.BOR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  `|Pr[GReal(Adv).main() @ &m : res] - Pr[G4.main() @ &m : res]| <=
   2%r * Switching.collBound.
proof.
intros LossPre LossPost.
cut GReal_G3 :
  `|Pr[GReal(Adv).main() @ &m : res] - Pr[G3.main() @ &m : res]| <=
   Switching.collBound.
  apply (GReal_G3 &m); [apply LossPre | apply LossPost].
cut G4_G3 :
  `|Pr[G4.main() @ &m : res] - Pr[G3.main() @ &m : res]| <=
   Switching.collBound.
  apply (G4_G3 &m); [apply LossPre | apply LossPost].
apply
  (Real.Trans
   _
   (`|Pr[GReal(Adv).main() @ &m : res] - Pr[G3.main() @ &m : res]| +
    `|Pr[G4.main() @ &m : res] - Pr[G3.main() @ &m : res]|)
   _).
apply
  (Triangle.triangular_inequality
   Pr[GReal(Adv).main() @ &m : res]
   Pr[G4.main() @ &m : res]
   Pr[G3.main() @ &m : res]).
apply (Real.Trans _ (Switching.collBound + Switching.collBound) _).
smt.
cut -> :
  Switching.collBound + Switching.collBound =
  2%r * Switching.collBound;
  first smt.
trivial.
qed.

(* In G5:

  * hdb is no longer a global variable;

  * serverHashAndCountDB is renamed to hashAndCountDB, takes in the
    list of queries as well as the database, only counts the numbers
    of queries (as opposed to arbitrary elements) in the database, but
    still hashes the database

  * apart from in the adversary's pre procedure, all hashing is done
    using hash (not bhash) *)

local module G5 = {
  module Or  = RO.Or
  module BOr = BRO.BOr(Or)

  var view : client_view
  var sec : sec

  proc hashAndCountDB(qrys : qrys, db : db) : elems_counts = {
    var qrysCounts : elems_counts;
    var i : int;
    var elem : elem;
    var tag : tag;
    qrysCounts = emptyEC;
    db = Shuffle.shuffle(db);
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      tag = BOr.hash((elem, sec));
      elem = oget(nth db i);
      if (mem elem qrys) {
        qrysCounts = incrCount qrysCounts elem;
      }
      i = i + 1;
    }
    return qrysCounts;
  }

  proc clientLoop(qrys : qrys, qrysCounts : elems_counts) : unit = {
    var i, n : int;
    var tag : tag;
    var qry : elem;
    view = view ++ [ClientViewQrys qrys; ClientViewGotSec sec];
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = BOr.hash((qry, sec));
      n = getCount qrysCounts qry;
      view = view ++ [ClientViewQryCount(qry, tag, n)];
      i = i + 1;
    }
  }

  module A = Adv(BOr)

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b, suff : bool;
    var qrysCounts : elems_counts;
    view = [];
    BOr.init();
    sec = $secDistr;
    (db, qrys) = A.pre(sec);
    suff = BOr.suff(length db  + length qrys);
    if (suff) {
      qrysCounts = hashAndCountDB(qrys, db);
      clientLoop(qrys, qrysCounts);
    }
    b = A.post(view);
    return b;
  }
}.

(* to make the transition from G4 to G5, we need this predicate: *)

pred ElemsToQrysCounts
     (qrys : qrys, elemsCounts qrysCounts : elems_counts) =
  forall (elem : elem),
  mem elem qrys =>
  (mem elem (dom elemsCounts) <=> mem elem (dom qrysCounts)) /\
  (mem elem (dom elemsCounts) => mem elem (dom qrysCounts) =>
   oget elemsCounts.[elem] = oget qrysCounts.[elem]).

local lemma ElemsToQrysCounts_Intro
  (qrys : qrys, elemsCounts qrysCounts : elems_counts) :
  (forall (elem : elem),
   mem elem qrys =>
   (mem elem (dom elemsCounts) <=> mem elem (dom qrysCounts)) /\
   (mem elem (dom elemsCounts) => mem elem (dom qrysCounts) =>
    oget elemsCounts.[elem] = oget qrysCounts.[elem])) =>
  ElemsToQrysCounts qrys elemsCounts qrysCounts
by trivial.

local lemma ElemsToQrysCounts_Elim
  (qrys : qrys, elemsCounts qrysCounts : elems_counts) :
  ElemsToQrysCounts qrys elemsCounts qrysCounts =>
  (forall (elem : elem),
   mem elem qrys =>
   (mem elem (dom elemsCounts) <=> mem elem (dom qrysCounts)) /\
   (mem elem (dom elemsCounts) => mem elem (dom qrysCounts) =>
    oget elemsCounts.[elem] = oget qrysCounts.[elem]))
by smt.

local lemma ElemsToQrysCounts_Empty (qrys : qrys) :
  ElemsToQrysCounts qrys emptyEC emptyEC
by trivial.

local lemma ElemsToQrysCounts_Qry
  (qrys : qrys, elem : elem, elemsCounts qrysCounts : elems_counts) :
  ElemsToQrysCounts qrys elemsCounts qrysCounts =>
  mem elem qrys =>
  ElemsToQrysCounts
  qrys
  (incrCount elemsCounts elem)
  (incrCount qrysCounts elem).
proof.
intros EC H1.
apply ElemsToQrysCounts_Intro.
intros elem0 H2.
case (elem0 = elem) => H3.
split.
cut -> : dom(incrCount elemsCounts elem) = add elem (dom elemsCounts) by smt.
cut -> : dom(incrCount qrysCounts elem) = add elem (dom qrysCounts) by smt.
smt.
intros H4 H5.
rewrite /incrCount; smt.
(* elem0 <> elem *)
split.
cut -> : dom(incrCount elemsCounts elem) = add elem (dom elemsCounts) by smt.
cut -> : dom(incrCount qrysCounts elem) = add elem (dom qrysCounts) by smt.
smt.
intros H4 H5.
cut H6 : mem elem0 (dom elemsCounts) <=> mem elem0 (dom qrysCounts); smt.
qed.

local lemma ElemsToQrysCounts_NonQry
  (qrys : qrys, elem : elem, elemsCounts qrysCounts : elems_counts) :
  ElemsToQrysCounts qrys elemsCounts qrysCounts =>
  ! mem elem qrys =>
  ElemsToQrysCounts qrys (incrCount elemsCounts elem) qrysCounts.
proof.
intros EC H1.
apply ElemsToQrysCounts_Intro.
intros elem0 H3.
rewrite /incrCount.
smt.
qed.

local lemma G4_G5_serverHashAndCountDB_hashAndCountDB (qrys' : qrys) :
  equiv
  [G4.serverHashAndCountDB ~ G5.hashAndCountDB :
   ={db, RO.Or.mp} /\ qrys{2} = qrys' /\ G4.sec{1} = G5.sec{2} ==>
   ={RO.Or.mp} /\ ElemsToQrysCounts qrys' res{1} res{2}].
proof.
proc.
seq 3 3 :
  (={RO.Or.mp, db, i} /\ G4.sec{1} = G5.sec{2} /\ qrys{2} = qrys' /\
  ElemsToQrysCounts qrys' elemsCounts{1} qrysCounts{2}).
wp; sp.
call (_ : ={xs} ==> ={res}).
sim.
skip; smt.
while
  (={db, RO.Or.mp, i} /\ G4.sec{1} = G5.sec{2} /\ qrys{2} = qrys' /\
   ElemsToQrysCounts qrys' elemsCounts{1} qrysCounts{2}).
wp.
call BRO.BOr_Or_bhash_hash.
auto; smt.
auto; smt.
qed.

local lemma G4_G5_clientLoop :
  equiv
  [G4.clientLoop ~ G5.clientLoop :
   ={qrys, RO.Or.mp} /\ G4.sec{1} = G5.sec{2} /\ G4.view{1} = G5.view{2} /\
   ElemsToQrysCounts qrys{1} elemsCounts{1} qrysCounts{2} ==>
   ={RO.Or.mp} /\ G4.view{1} = G5.view{2}].
proof.
proc.
while
  (={qrys, RO.Or.mp, i} /\ G4.sec{1} = G5.sec{2} /\
   G4.view{1} = G5.view{2} /\
   ElemsToQrysCounts qrys{1} elemsCounts{1} qrysCounts{2} /\
   i{1} >= 0).
wp.
call BRO.BOr_Or_bhash_hash.
auto; progress; smt.
auto; smt.
qed.

(* now we show the equivalence of G4 and G5: *)

local lemma G4_G5_main :
  equiv[G4.main ~ G5.main : true ==> ={res}].
proof.
proc.
(* initialization and calls to pre and suff *)
seq 5 5 :
  (={db, qrys, suff, glob BRO.BOr(RO.Or), glob Adv} /\
   G4.sec{1} = G5.sec{2} /\ G4.view{1} = G5.view{2}).
sim.
(* now handle if *)
if => //.
(* then branch *)
call (_ : ={view, RO.Or.mp, glob Adv} ==> ={res}).
proc (={RO.Or.mp}); trivial.
sim.
call G4_G5_clientLoop.
exists* qrys{1}; elim*; intros qrys_L.
call (G4_G5_serverHashAndCountDB_hashAndCountDB qrys_L).
auto.
(* else branch *)
call (_ : ={view, RO.Or.mp, glob Adv} ==> ={res}).
proc (={RO.Or.mp}); trivial.
sim.
auto.
qed.

local lemma G4_G5 &m :
  Pr[G4.main() @ &m : res] = Pr[G5.main() @ &m : res].
proof.
byequiv G4_G5_main; trivial.
qed.

local lemma GReal_G5 &m :
  (forall (O <: BRO.BOR{Adv}),
   islossless O.bhash => islossless Adv(O).pre) =>
  (forall (O <: BRO.BOR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  `|Pr[GReal(Adv).main() @ &m : res] - Pr[G5.main() @ &m : res]| <=
   2%r * Switching.collBound.
proof.
intros LossPre LossPost.
rewrite - (G4_G5 &m).
apply (GReal_G4 &m); [apply LossPre | apply LossPost].
qed.

(* in G6, we prepare the while loop in hashAndCountDB for list
   fission *)

local module G6 = {
  module Or  = RO.Or
  module BOr = BRO.BOr(Or)

  var view : client_view
  var sec : sec

  proc hashAndCountDB(qrys : qrys, db : db) : elems_counts = {
    var qrysCounts : elems_counts;
    var i : int;
    var elem, elem' : elem;
    var tag : tag;
    qrysCounts = emptyEC;
    db = Shuffle.shuffle(db);
    i = 0;
    while (i < length db) {
      (* hashing *)
      elem = oget(nth db i);
      tag = BOr.hash((elem, sec));
      (* counting *)
      elem' = oget(nth db i);
      if (mem elem' qrys) {
        qrysCounts = incrCount qrysCounts elem';
      }
      i = i + 1;
    }
    return qrysCounts;
  }

  proc clientLoop(qrys : qrys, qrysCounts : elems_counts) : unit = {
    var i, n : int;
    var tag : tag;
    var qry : elem;
    view = view ++ [ClientViewQrys qrys; ClientViewGotSec sec];
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = BOr.hash((qry, sec));
      n = getCount qrysCounts qry;
      view = view ++ [ClientViewQryCount(qry, tag, n)];
      i = i + 1;
    }
  }

  module A = Adv(BOr)

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b, suff : bool;
    var qrysCounts : elems_counts;
    view = [];
    BOr.init();
    sec = $secDistr;
    (db, qrys) = A.pre(sec);
    suff = BOr.suff(length db  + length qrys);
    if (suff) {
      qrysCounts = hashAndCountDB(qrys, db);
      clientLoop(qrys, qrysCounts);
    }
    b = A.post(view);
    return b;
  }
}.

local lemma G5_G6_hashAndCountDB :
  equiv
  [G5.hashAndCountDB ~ G6.hashAndCountDB :
   ={qrys, db, RO.Or.mp} /\ G5.sec{1} = G6.sec{2} ==>
   ={res, RO.Or.mp}].
proof.
proc.
while (={i, qrys, db, qrysCounts, RO.Or.mp} /\ G5.sec{1} = G6.sec{2}).
wp.
call (_ : ={inp, RO.Or.mp} ==> ={res, RO.Or.mp}).
sim.
auto.
wp; sp.
call (_ : ={xs} ==> ={res}); first sim.
auto.
qed.

local lemma G5_G6_clientLoop :
  equiv
  [G5.clientLoop ~ G6.clientLoop :
   ={qrys, qrysCounts, RO.Or.mp} /\ G5.sec{1} = G6.sec{2} /\
   G5.view{1} = G6.view{2} ==>
   ={RO.Or.mp} /\ G5.view{1} = G6.view{2}].
proof.
sim.
qed.

local lemma G5_G6_main :
  equiv[G5.main ~ G6.main : true ==> ={res}].
proof.
proc.
(* initialization and call to pre *)
seq 5 5 :
  (={qrys, db, suff, RO.Or.mp, glob Adv} /\ G5.sec{1} = G6.sec{2} /\
   G5.view{1} = G6.view{2}).
sim.
(* handle if *)
if => //.
(* then branch *)
call (_ : ={view, RO.Or.mp, glob Adv} ==> ={res}).
proc (={RO.Or.mp}) => //.
sim.
call G5_G6_clientLoop.
call G5_G6_hashAndCountDB.
auto.
(* else branch *)
call (_ : ={view, RO.Or.mp, glob Adv} ==> ={res}).
proc (={RO.Or.mp}) => //.
sim.
auto.
qed.

local lemma G5_G6 &m :
  Pr[G5.main() @ &m : res] = Pr[G6.main() @ &m : res].
proof.
byequiv G5_G6_main; trivial.
qed.

local lemma GReal_G6 &m :
  (forall (O <: BRO.BOR{Adv}),
   islossless O.bhash => islossless Adv(O).pre) =>
  (forall (O <: BRO.BOR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  `|Pr[GReal(Adv).main() @ &m : res] - Pr[G6.main() @ &m : res]| <=
   2%r * Switching.collBound.
proof.
intros LossPre LossPost.
rewrite - (G5_G6 &m).
apply (GReal_G5 &m); [apply LossPre | apply LossPost].
qed.

(* in G7, we carry out loop fission in countAndHashDB, splitting
   the result into one function for doing the hashing, and another
   for doing the counting *)

local module G7 = {
  module Or  = RO.Or
  module BOr = BRO.BOr(Or)

  var view : client_view
  var sec : sec

  proc hashDB(db : db) : unit = {
    var i : int;
    var elem : elem;
    var tag : tag;
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      tag = BOr.hash((elem, sec));
      i = i + 1;
    }
  }

  proc computeAnswers(qrys : qrys, db : db) : elems_counts = {
    var qrysCounts : elems_counts;
    var i : int;
    var elem : elem;
    qrysCounts = emptyEC;
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      if (mem elem qrys) {
        qrysCounts = incrCount qrysCounts elem;
      }
      i = i + 1;
    }
    return qrysCounts;
  }

  proc clientLoop(qrys : qrys, qrysCounts : elems_counts) : unit = {
    var i, n : int;
    var tag : tag;
    var qry : elem;
    view = view ++ [ClientViewQrys qrys; ClientViewGotSec sec];
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = BOr.hash((qry, sec));
      n = getCount qrysCounts qry;
      view = view ++ [ClientViewQryCount(qry, tag, n)];
      i = i + 1;
    }
  }

  module A = Adv(BRO.BOr(RO.Or))

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b, suff : bool;
    var qrysCounts : elems_counts;
    view = [];
    BRO.BOr(RO.Or).init();
    sec = $secDistr;
    (db, qrys) = A.pre(sec);
    suff = BOr.suff(length db  + length qrys);
    if (suff) {
      db = Shuffle.shuffle(db);
      qrysCounts = computeAnswers(qrys, db);
      hashDB(db);
      clientLoop(qrys, qrysCounts);
    }
    b = A.post(view);
    return b;
  }
}.

local lemma G6_G7_main :
  equiv[G6.main ~ G7.main : true ==> ={res}].
proof.
proc.
(* initialization and call to pre *)
seq 5 5 :
  (={qrys, db, suff, RO.Or.mp, glob Adv} /\ G6.sec{1} = G7.sec{2} /\
   G6.view{1} = G7.view{2}).
sim.
(* handle if *)
if => //.
(* then branch *)
call (_ : ={view, RO.Or.mp, glob Adv} ==> ={res}).
proc (={RO.Or.mp}) => //.
sim.
call
  (_ :
   ={qrys, qrysCounts, RO.Or.mp} /\ G6.sec{1} = G7.sec{2} /\
   G6.view{1} = G7.view{2} ==>
   ={res, RO.Or.mp} /\ G6.view{1} = G7.view{2}).
sim.
swap {2} 3 -1.
inline G6.hashAndCountDB G7.hashDB G7.computeAnswers.
fission {1} 6 @ 2, 4.
swap {1} 3 3.
swap {1} 1 6.
(* handle hashing *)
seq 4 4 :
  (={qrys, db0, suff, RO.Or.mp, glob Adv} /\ db{2} = db0{2} /\
   G6.sec{1} = G7.sec{2} /\ G6.view{1} = G7.view{2}).
while (={i, db0, RO.Or.mp} /\ G6.sec{1} = G7.sec{2}).
wp.
call (_ : ={inp, RO.Or.mp} ==> ={res, RO.Or.mp}).
sim.
auto.
wp; sp.
call (_ : ={xs} ==> ={res}); first sim.
auto.
(* handle counting *)
wp.
while
  (={RO.Or.mp, qrysCounts0} /\ i{1} = i0{2} /\ db0{1} = db1{2} /\
   qrys0{1} = qrys0{2} /\ i{1} <= length(db0{1})).
auto; smt.
auto; smt.
(* else branch *)
sim.
qed.

local lemma G6_G7 &m :
  Pr[G6.main() @ &m : res] = Pr[G7.main() @ &m : res].
proof.
byequiv G6_G7_main; trivial.
qed.

local lemma GReal_G7 &m :
  (forall (O <: BRO.BOR{Adv}),
   islossless O.bhash => islossless Adv(O).pre) =>
  (forall (O <: BRO.BOR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  `|Pr[GReal(Adv).main() @ &m : res] - Pr[G7.main() @ &m : res]| <=
   2%r * Switching.collBound.
proof.
intros LossPre LossPost.
rewrite - (G6_G7 &m).
apply (GReal_G6 &m); [apply LossPre | apply LossPost].
qed.

(* in G8, we get rid of the call to hashDB *)

local module G8 = {
  module Or  = RO.Or
  module BOr = BRO.BOr(Or)

  var view : client_view
  var sec : sec

  proc computeAnswers(qrys : qrys, db : db) : elems_counts = {
    var qrysCounts : elems_counts;
    var i : int;
    var elem : elem;
    qrysCounts = emptyEC;
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      if (mem elem qrys) {
        qrysCounts = incrCount qrysCounts elem;
      }
      i = i + 1;
    }
    return qrysCounts;
  }

  proc clientLoop(qrys : qrys, qrysCounts : elems_counts) : unit = {
    var i, n : int;
    var tag : tag;
    var qry : elem;
    view = view ++ [ClientViewQrys qrys; ClientViewGotSec sec];
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = BOr.hash((qry, sec));
      n = getCount qrysCounts qry;
      view = view ++ [ClientViewQryCount(qry, tag, n)];
      i = i + 1;
    }
  }

  module A = Adv(BRO.BOr(RO.Or))

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b, suff : bool;
    var qrysCounts : elems_counts;
    view = [];
    BRO.BOr(RO.Or).init();
    sec = $secDistr;
    (db, qrys) = A.pre(sec);
    suff = BOr.suff(length db  + length qrys);
    if (suff) {
      db = Shuffle.shuffle(db);
      qrysCounts = computeAnswers(qrys, db);
      clientLoop(qrys, qrysCounts);
    }
    b = A.post(view);
    return b;
  }
}.

(* we locally clone the abstract theory BRO.RedundantHashing as the
   theory RH, fully realizing it (there is nothing to realize) *)

local clone BRO.RedundantHashing as RH
proof *.
(* nothing to realize *)

(* we split G7 into a redundant hashing adversary RHAdv plus a hashing
   module Hashing, so that RH.GDoHashing(RHAdv, Hashing) is equivalent
   to G7, and RH.GNoHashing(RHAdv) is equivalent to G8 *)

local module RHAdv(O : BRO.BOR) : RH.RHADV(O) = {
  (* globals for communicating between pre/post and the procedures
     they call *)
  var view : client_view
  var sec : sec
  var db : db
  var qrys : qrys
  var qrysCounts : elems_counts

  (* called by pre *)
  proc computeAnswers(qrys : qrys, db : db) : elems_counts = {
    var qrysCounts : elems_counts;
    var i : int;
    var elem : elem;
    qrysCounts = emptyEC;
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      if (mem elem qrys) {
        qrysCounts = incrCount qrysCounts elem;
      }
      i = i + 1;
    }
    return qrysCounts;
  }

  (* called by post *)
  proc clientLoop(qrys : qrys, qrysCounts : elems_counts) : unit = {
    var i, n : int;
    var tag : tag;
    var qry : elem;
    view = view ++ [ClientViewQrys qrys; ClientViewGotSec sec];
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = O.hash((qry, sec));
      n = getCount qrysCounts qry;
      view = view ++ [ClientViewQryCount(qry, tag, n)];
      i = i + 1;
    }
  }

  module A = Adv(O)

  proc pre() : RH.rh_pre_res = {
    var suff, b : bool;
    var rhpr : RH.rh_pre_res;

    view = [];
    sec = $secDistr;
    (db, qrys) = A.pre(sec);
    suff = O.suff(length db  + length qrys);
    qrysCounts = emptyEC;  (* need unconditional initialization *)
    if (suff) {
      db = Shuffle.shuffle(db);
      qrysCounts = computeAnswers(qrys, db);
      rhpr = Right(map (fun elem, (elem, sec)) db);
    }
    else {
      b = A.post(view);
      rhpr = Left b;
    }
    return rhpr;
  }

  proc post() : bool = {
    var b : bool;

    clientLoop(qrys, qrysCounts);
    b = A.post(view);
    return b;
  }
}.

local module Hashing(O : BRO.BOR) : RH.HASHING(O) = {
  proc hashing(inps : (elem * sec)list) : unit = {
    var i : int;
    var inp : elem * sec;
    var tag : tag;
    i = 0;
    while (i < length inps) {
      inp = oget(nth inps i);
      tag = O.hash(inp);
      i = i + 1;
    }
  }
}.

local lemma Hashing_Lossless_hashing :
  forall (O <: BRO.BOR{Hashing}),
  islossless O.hash => islossless Hashing(O).hashing.
proof.
intros O LossHash.
proc.
while (true) (length inps - i).
auto.
call (_ : true ==> true) => //.
auto; smt.
auto; smt.
qed.

local lemma G7_GDoHashing_RHAdv_Hashing_main :
  equiv[G7.main ~ RH.GDoHashing(RHAdv, Hashing).main : true ==> ={res}].
proc.
inline
  RH.GDoHashing(RHAdv, Hashing).RHA.pre
  RH.GDoHashing(RHAdv, Hashing).RHA.post.
swap {1} 1.
seq 5 6 :
  (={glob BRO.BOr(RO.Or), glob Adv, suff} /\ G7.sec{1} = RHAdv.sec{2} /\
   G7.view{1} = RHAdv.view{2} /\ db{1} = RHAdv.db{2} /\
   qrys{1} = RHAdv.qrys{2}).
sim.
if => //.
(* then branch *)
seq 2 2 :
  (={glob BRO.BOr(RO.Or), glob Adv, suff} /\ G7.sec{1} = RHAdv.sec{2} /\
   G7.view{1} = RHAdv.view{2} /\ db{1} = RHAdv.db{2} /\
   qrys{1} = RHAdv.qrys{2} /\
   qrysCounts{1} = RHAdv.qrysCounts{2}); first sim.
seq 0 2 :
  (={glob BRO.BOr(RO.Or), glob Adv, suff} /\ G7.sec{1} = RHAdv.sec{2} /\
   G7.view{1} = RHAdv.view{2} /\ db{1} = RHAdv.db{2} /\
   qrys{1} = RHAdv.qrys{2} /\ qrysCounts{1} = RHAdv.qrysCounts{2} /\
   rhpr{2} = Right(map (fun elem, (elem, RHAdv.sec{2})) RHAdv.db{2}));
  first auto.
rcondf {2} 1; first auto.
inline G7.hashDB RH.GDoHashing(RHAdv, Hashing).H.hashing.
seq 1 2 :
  (={glob BRO.BOr(RO.Or), glob Adv, suff} /\ G7.sec{1} = RHAdv.sec{2} /\
   G7.view{1} = RHAdv.view{2} /\ db{1} = RHAdv.db{2} /\
   qrys{1} = RHAdv.qrys{2} /\ qrysCounts{1} = RHAdv.qrysCounts{2} /\
   inps0{2} = map (fun elem, (elem, RHAdv.sec{2})) db0{1} /\
   length db0{1} = length inps0{2}).
auto; smt.
sim.
while
  (={i, glob BRO.BOr(RO.Or)} /\ 0 <= i{1} <= length db0{1} /\
   G7.sec{1} = RHAdv.sec{2} /\ length db0{1} = length inps0{2} /\
   inps0{2} = map (fun elem, (elem, RHAdv.sec{2})) db0{1}).
wp.
call (_ : ={inp, RO.Or.mp} ==> ={res, RO.Or.mp}).
sim.
auto; smt.
auto; smt.
(* else branch *)
seq 1 3 :
  (={glob BRO.BOr(RO.Or), glob Adv} /\ b{1} = b0{2} /\ rhpr{2} = Left b0{2}).
wp.
call
  (_ :
   ={view, glob BRO.BOr(RO.Or), glob Adv} ==>
   ={res, glob BRO.BOr(RO.Or), glob Adv});
  first sim.
auto.
rcondt {2} 1; first auto.
auto.
qed.

local lemma G7_GDoHashing_RHAdv_Hashing &m :
  Pr[G7.main() @ &m : res] =
  Pr[RH.GDoHashing(RHAdv, Hashing).main() @ &m : res].
proof.
byequiv G7_GDoHashing_RHAdv_Hashing_main => //.
qed.

local lemma GNoHashing_RHAdv_G8_main :
  equiv[RH.GNoHashing(RHAdv).main ~ G8.main : true ==> ={res}].
proc.
inline RH.GNoHashing(RHAdv).RHA.pre RH.GNoHashing(RHAdv).RHA.post.
seq 6 5 :
  (={glob BRO.BOr(RO.Or), glob Adv, suff} /\ RHAdv.sec{1} = G8.sec{2} /\
   RHAdv.view{1} = G8.view{2} /\ RHAdv.view{1} = G8.view{2} /\
   RHAdv.db{1} = db{2} /\ RHAdv.qrys{1} = qrys{2}).
swap {1} 1; sim.
if => //.
(* then branch *)
seq 2 2 :
  (={glob BRO.BOr(RO.Or), glob Adv, suff} /\ RHAdv.sec{1} = G8.sec{2} /\
   RHAdv.view{1} = G8.view{2} /\ RHAdv.db{1} = db{2} /\
   RHAdv.qrys{1} = qrys{2} /\ RHAdv.qrysCounts{1} = qrysCounts{2});
  first sim.
seq 2 0 :
  (={glob BRO.BOr(RO.Or), glob Adv, suff} /\ RHAdv.sec{1} = G8.sec{2} /\
   RHAdv.view{1} = G8.view{2} /\ RHAdv.db{1} = db{2} /\
   RHAdv.qrys{1} = qrys{2} /\ RHAdv.qrysCounts{1} = qrysCounts{2} /\
   rhpr{1} = Right(map (fun elem, (elem, RHAdv.sec{1})) RHAdv.db{1}));
  first auto.
rcondf {1} 1; first auto.
sim.
(* else branch *)
seq 3 1 :
  (={glob BRO.BOr(RO.Or), glob Adv} /\ b0{1} = b{2} /\ rhpr{1} = Left b0{1}).
wp; sp.
call
  (_ :
   ={view, glob BRO.BOr(RO.Or), glob Adv} ==>
   ={res, glob BRO.BOr(RO.Or), glob Adv});
  first sim.
auto.
rcondt {1} 1; first auto.
auto.
qed.

local lemma GNoHashing_RHAdv_G8 &m :
  Pr[RH.GNoHashing(RHAdv).main() @ &m : res] =
  Pr[G8.main() @ &m : res].
proof.
byequiv GNoHashing_RHAdv_G8_main => //.
qed.

(* we now apply RH.GDoHashing_GNoHashing to RHAdv and Hashing: *)

local lemma G7_G8 &m :
  Pr[G7.main() @ &m : res] = Pr[G8.main() @ &m : res].
proof.
rewrite (G7_GDoHashing_RHAdv_Hashing &m).
rewrite - (GNoHashing_RHAdv_G8 &m).
apply (RH.GDoHashing_GNoHashing RHAdv Hashing &m).
apply Hashing_Lossless_hashing.
qed.

local lemma GReal_G8 &m :
  (forall (O <: BRO.BOR{Adv}),
   islossless O.bhash => islossless Adv(O).pre) =>
  (forall (O <: BRO.BOR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  `|Pr[GReal(Adv).main() @ &m : res] - Pr[G8.main() @ &m : res]| <=
   2%r * Switching.collBound.
proof.
intros LossPre LossPost.
rewrite - (G7_G8 &m).
apply (GReal_G7 &m); [apply LossPre | apply LossPost].
qed.

(* in G9, we get rid of the database shuffling that preceded the call
   to computeAnswers *)

local module G9 = {
  module Or  = RO.Or
  module BOr = BRO.BOr(Or)

  var view : client_view
  var sec : sec

  proc computeAnswers(qrys : qrys, db : db) : elems_counts = {
    var qrysCounts : elems_counts;
    var i : int;
    var elem : elem;
    qrysCounts = emptyEC;
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      if (mem elem qrys) {
        qrysCounts = incrCount qrysCounts elem;
      }
      i = i + 1;
    }
    return qrysCounts;
  }

  proc clientLoop(qrys : qrys, qrysCounts : elems_counts) : unit = {
    var i, n : int;
    var tag : tag;
    var qry : elem;
    view = view ++ [ClientViewQrys qrys; ClientViewGotSec sec];
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = BOr.hash((qry, sec));
      n = getCount qrysCounts qry;
      view = view ++ [ClientViewQryCount(qry, tag, n)];
      i = i + 1;
    }
  }

  module A = Adv(BRO.BOr(RO.Or))

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b, suff : bool;
    var qrysCounts : elems_counts;
    view = [];
    BRO.BOr(RO.Or).init();
    sec = $secDistr;
    (db, qrys) = A.pre(sec);
    suff = BOr.suff(length db  + length qrys);
    if (suff) {
      qrysCounts = computeAnswers(qrys, db);
      clientLoop(qrys, qrysCounts);
    }
    b = A.post(view);
    return b;
  }
}.

pred NumOccsListEqual (xs ys : elem list) =
  (forall (elem : elem),
   numOccs elem xs = numOccs elem ys) /\
  `|xs| = `|ys|.  (* redundant, but less work to include *)

pred NumOccsListMapEqual
     (i : int) (qrys db1 db2 : elem list)
     (qrysCounts1 qrysCounts2 : (elem, int)map) =
  `|db1| = `|db2| /\ 0 <= i <= `|db1| /\
  (forall (elem : elem),
   mem elem (dom qrysCounts1) =>
   mem elem qrys /\ oget(qrysCounts1.[elem]) >= 1) /\
  (forall (elem : elem),
   mem elem (dom qrysCounts2) =>
   mem elem qrys /\ oget(qrysCounts2.[elem]) >= 1) /\
  (forall (elem : elem),
   mem elem qrys =>
   numOccsUpTo elem (rev db1) (`|db1| - i) +
   getCount qrysCounts1 elem =
   numOccsUpTo elem (rev db2) (`|db2| - i) +
   getCount qrysCounts2 elem).

lemma NumOccsListMapEqualInit (qrys db1 db2 : elem list) :
  NumOccsListEqual db1 db2 =>
  NumOccsListMapEqual 0 qrys db1 db2 emptyEC emptyEC.
proof.
intros H1.
split; first smt.
split; first smt.
split; first smt.
split; first smt.
intros elem H2.
simplify.
cut -> : getCount emptyEC elem = 0 by smt.
cut -> : `|db1| = `|rev db1| by smt.
cut -> : `|db2| = `|rev db2| by smt.
simplify.
cut -> :
  numOccsUpTo elem (rev db1) (`|rev db1|) =
  numOccs elem (rev db1) by smt.
cut -> :
  numOccsUpTo elem (rev db2) (`|rev db2|) =
  numOccs elem (rev db2) by smt.
rewrite NumOccsRev.
rewrite NumOccsRev.
smt.
qed.

lemma NumOccsListMapEqualFinal
      (qrys db1 db2 : elem list) (qrysCounts1 qrysCounts2 : (elem, int)map) :
  NumOccsListMapEqual `|db1| qrys db1 db2 qrysCounts1 qrysCounts2 =>
  qrysCounts1 = qrysCounts2.
proof.
intros H1.
apply map_ext.
intros elem.
elim H1 => H1 [H2 [H3 [H4 H5]]].
cut H6 : 
  numOccsUpTo elem (rev db1) (`|db1| - `|db1|) +
  getCount qrysCounts1 elem =
  numOccsUpTo elem (rev db2) (`|db2| - `|db1|) +
  getCount qrysCounts2 elem by smt.
cut H7 :
  getCount qrysCounts1 elem = getCount qrysCounts2 elem; smt.
qed.

lemma NumOccsListMapEqualStepNeitherQry
      (i : int) (qrys db1 db2 : elem list)
      (qrysCounts1 qrysCounts2 : (elem, int)map) :
  NumOccsListMapEqual i qrys db1 db2 qrysCounts1 qrysCounts2 =>
  0 <= i < `|db1| =>
  ! mem (oget(nth db1 i)) qrys /\ ! mem (oget(nth db2 i)) qrys =>
  NumOccsListMapEqual (i + 1) qrys db1 db2 qrysCounts1 qrysCounts2.
proof.
intros [H1 [H2 [H3 [H4 H5]]]] H6 H7.
split; first trivial.
split; first smt.
split; first smt.
split; first smt.
intros elem H8.
cut H9 :
  numOccsUpTo elem (rev db1) (`|db1| - i) + getCount qrysCounts1 elem =
  numOccsUpTo elem (rev db2) (`|db2| - i) + getCount qrysCounts2 elem;
  smt.
qed.

lemma NumOccsListMapEqualStepOnlyFirstQry
      (i : int) (qrys db1 db2 : elem list)
      (qrysCounts1 qrysCounts2 : (elem, int)map) :
  NumOccsListMapEqual i qrys db1 db2 qrysCounts1 qrysCounts2 =>
  0 <= i < `|db1| =>
  mem (oget(nth db1 i)) qrys /\ !(mem (oget(nth db2 i)) qrys) =>
  NumOccsListMapEqual (i + 1) qrys db1 db2
  (incrCount qrysCounts1 (oget(nth db1 i))) qrysCounts2.
proof.
intros [H1 [H2 [H3 [H4 H5]]]] H6 H7.
split; first trivial.
split; first smt.
split.
intros elem H8.
apply (ECInvar qrysCounts1 (oget(nth db1 i)) qrys); smt.
split; first smt.
(* last conjunct of NumOccsListMapEqual *)
intros elem H8.
cut H9 : elem <> oget(nth db2 i) by smt.
cut -> :
  numOccsUpTo elem (rev db2) (`|db2| - (i + 1)) =
  numOccsUpTo elem (rev db2) (`|db2| - i) by smt.
case (elem = oget(nth db1 i)) => H10.
cut -> :
  numOccsUpTo elem (rev db1) (`|db1| - (i + 1)) =
  numOccsUpTo elem (rev db1) (`|db1| - i) - 1 by smt.
cut -> :
  getCount (incrCount qrysCounts1 (oget(nth db1 i))) elem =
  getCount qrysCounts1 elem + 1; smt.
(* elem <> oget(nth db1 i) *)
cut -> :
  numOccsUpTo elem (rev db1) (`|db1| - (i + 1)) =
  numOccsUpTo elem (rev db1) (`|db1| - i) by smt.
cut -> :
  getCount (incrCount qrysCounts1 (oget(nth db1 i))) elem =
  getCount qrysCounts1 elem; smt.
qed.

lemma NumOccsListMapEqualStepOnlySndQry
      (i : int) (qrys db1 db2 : elem list)
      (qrysCounts1 qrysCounts2 : (elem, int)map) :
  NumOccsListMapEqual i qrys db1 db2 qrysCounts1 qrysCounts2 =>
  i < `|db1| =>
  !(mem (oget(nth db1 i)) qrys) /\ mem (oget(nth db2 i)) qrys =>
  NumOccsListMapEqual (i + 1) qrys db1 db2
  qrysCounts1 (incrCount qrysCounts2 (oget(nth db2 i))).
proof.
intros [H1 [H2 [H3 [H4 H5]]]] H6 H7.
split; first trivial.
split; first smt.
split; first smt.
split.
intros elem H8.
apply ECInvar; smt.
(* last conjunct of NumOccsListMapEqual *)
intros elem H8.
cut H9 : elem <> oget(nth db1 i) by smt.
cut -> :
  numOccsUpTo elem (rev db1) (`|db1| - (i + 1)) =
  numOccsUpTo elem (rev db1) (`|db1| - i) by smt.
case (elem = oget(nth db2 i)) => H10.
cut -> :
  numOccsUpTo elem (rev db2) (`|db2| - (i + 1)) =
  numOccsUpTo elem (rev db2) (`|db2| - i) - 1 by smt.
cut -> :
  getCount (incrCount qrysCounts2 (oget (nth db2 i))) elem =
  getCount qrysCounts2 elem + 1; smt.
(* elem <> oget(nth db2 i) *)
cut -> :
  numOccsUpTo elem (rev db2) (`|db2| - (i + 1)) =
  numOccsUpTo elem (rev db2) (`|db2| - i) by smt.
cut -> :
  getCount (incrCount qrysCounts2 (oget (nth db2 i))) elem =
  getCount qrysCounts2 elem; smt.
qed.

lemma NumOccsListMapEqualStepOnlyBothQrys
      (i : int) (qrys db1 db2 : elem list)
      (qrysCounts1 qrysCounts2 : (elem, int)map) :
  NumOccsListMapEqual i qrys db1 db2 qrysCounts1 qrysCounts2 =>
  i < `|db1| =>
  mem (oget(nth db1 i)) qrys /\ mem (oget(nth db2 i)) qrys =>
  NumOccsListMapEqual (i + 1) qrys db1 db2
  (incrCount qrysCounts1 (oget(nth db1 i)))
  (incrCount qrysCounts2 (oget(nth db2 i))).
proof.
intros [H1 [H2 [H3 [H4 H5]]]] H6 H7.
split; first trivial.
split; first smt.
split.
intros elem H8.
apply ECInvar; smt.
split.
intros elem H8.
apply ECInvar; smt.
intros elem H8.
case (elem = oget(nth db1 i)) => H9.
cut -> :
  numOccsUpTo elem (rev db1) (`|db1| - (i + 1)) =
  numOccsUpTo elem (rev db1) (`|db1| - i) - 1 by smt.
case (elem = oget(nth db2 i)) => H10.
cut -> :
  numOccsUpTo elem (rev db2) (`|db2| - (i + 1)) =
  numOccsUpTo elem (rev db2) (`|db2| - i) - 1 by smt.
cut -> :
  getCount (incrCount qrysCounts1 (oget(nth db1 i))) elem =
  getCount qrysCounts1 elem + 1 by smt.
cut -> :
  getCount (incrCount qrysCounts2 (oget(nth db2 i))) elem =
  getCount qrysCounts2 elem + 1; smt.
(* elem = oget(nth db1 i) /\ elem <> oget(nth db2 i) *)
cut -> :
  numOccsUpTo elem (rev db2) (`|db2| - (i + 1)) =
  numOccsUpTo elem (rev db2) (`|db2| - i) by smt.
cut -> :
  getCount (incrCount qrysCounts2 (oget(nth db2 i))) elem =
  getCount qrysCounts2 elem; smt.
(* elem <> oget(nth db1 i) *)
cut -> :
  numOccsUpTo elem (rev db1) (`|db1| - (i + 1)) =
  numOccsUpTo elem (rev db1) (`|db1| - i) by smt.
cut -> :
  getCount (incrCount qrysCounts1 (oget(nth db1 i))) elem =
  getCount qrysCounts1 elem by smt.
case (elem = oget(nth db2 i)) => H10.
(* elem <> oget(nth db1 i) /\ elem = oget(nth db2 i) *)
cut -> :
  numOccsUpTo elem (rev db2) (`|db2| - (i + 1)) =
  numOccsUpTo elem (rev db2) (`|db2| - i) - 1 by smt.
cut -> :
  getCount (incrCount qrysCounts2 (oget(nth db2 i))) elem =
  getCount qrysCounts2 elem + 1; smt.
(* elem <> oget(nth db1 i) /\ elem <> oget(nth db2 i) *)
cut -> :
  numOccsUpTo elem (rev db2) (`|db2| - (i + 1)) =
  numOccsUpTo elem (rev db2) (`|db2| - i) by smt.
cut -> :
  getCount (incrCount qrysCounts2 (oget(nth db2 i))) elem =
  getCount qrysCounts2 elem; smt.
qed.

local lemma G8_G9_computeAnswers :
  equiv
  [G8.computeAnswers ~ G9.computeAnswers :
   ={qrys} /\ NumOccsListEqual db{1} db{2} ==>
   ={res}].
proof.
proc.
(* initialization *)
seq 2 2 :
  (={qrys, i} /\ i{1} = 0 /\
   NumOccsListMapEqual i{1} qrys{1} db{1} db{2}
   qrysCounts{1} qrysCounts{2}).
auto; smt.
(* while loop *)
while
  (={i, qrys} /\
   NumOccsListMapEqual i{1} qrys{1} db{1} db{2} qrysCounts{1} qrysCounts{2}).
seq 1 1 :
  ((={i, qrys} /\ i{1} < length db{1} /\ i{2} < length db{2} /\
   NumOccsListMapEqual i{1} qrys{1} db{1} db{2} qrysCounts{1} qrysCounts{2}) /\
   elem{1} = oget(nth db{1} i{1}) /\ elem{2} = oget(nth db{2} i{2})).
auto; smt.
if {1}.
if {2}.
auto; smt.
auto; smt.
if {2}.
auto; smt.
auto; smt.
(* outside while loop *)
skip.
intros &1 &2 H1.
split; first smt.
intros qrysCounts1 i1 qrysCounts2 i2 H2 H3 H4.
cut H5 : i1 = `|length db{1}| by smt.
cut H6 : i2 = `|length db{2}| by smt.
cut H7 : i1 = i2 by smt.
apply
  (NumOccsListMapEqualFinal qrys{1} db{1} db{2}
   qrysCounts1 qrysCounts2).
smt.
qed.

local lemma G8_G9_main :
  equiv[G8.main ~ G9.main : true ==> ={res}].
proof.
proc.
(* initialization, call to pre, sufficiency check *)
seq 5 5 :
  (={db, qrys, suff, glob BRO.BOr(RO.Or), glob Adv} /\
   G8.sec{1} = G9.sec{2} /\ G8.view{1} = G9.view{2}).
sim.
(* conditional *)
if => //.
(* then part *)
(* handle shuffling in {1} *)
seq 1 0 :
  (={qrys, glob BRO.BOr(RO.Or), glob Adv} /\ G8.sec{1} = G9.sec{2} /\
   G8.view{1} = G9.view{2} /\ NumOccsListEqual db{1} db{2}).
inline Shuffle.shuffle.
wp; sp.
while {1}
  (`|xs{1}| + `|ys{1}| = `|db{2}| /\
   forall (elem : elem),
   numOccs elem xs{1} + numOccs elem ys{1} = numOccs elem db{2})
  (length xs{1}).
intros &m z.
wp.
rnd (fun i, 0 <= i <= length xs - 1).
skip.
intros &hr H1.
split.
apply Dinter.mu_in_supp; smt.
intros H2 v H3.
simplify.
cut -> : (0 <= v <= length xs{hr} - 1) = true by smt.
simplify.
split; [split; smt | smt].
skip; progress; smt.
(* handle call to computeAnswers *)
seq 1 1 :
  (={qrys, glob BRO.BOr(RO.Or), glob Adv, qrysCounts} /\
   G8.sec{1} = G9.sec{2} /\ G8.view{1} = G9.view{2}).
call
  (_ :
   ={qrys} /\ NumOccsListEqual db{1} db{2} ==>
   ={res}).
apply G8_G9_computeAnswers.
skip; smt.
(* client loop and post *)
sim.
(* else part *)
sim.
qed.

local lemma G8_G9 &m :
  Pr[G8.main() @ &m : res] = Pr[G9.main() @ &m : res].
proof.
byequiv G8_G9_main => //.
qed.

local lemma GReal_G9 &m :
  (forall (O <: BRO.BOR{Adv}),
   islossless O.bhash => islossless Adv(O).pre) =>
  (forall (O <: BRO.BOR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  `|Pr[GReal(Adv).main() @ &m : res] - Pr[G9.main() @ &m : res]| <=
   2%r * Switching.collBound.
proof.
intros LossPre LossPost.
rewrite - (G8_G9 &m).
apply (GReal_G8 &m); [apply LossPre | apply LossPost].
qed.

(* we get from GIdeal(Adv) to G9 by inlining *)

local lemma G9_GIdeal_main :
  equiv[G9.main ~ GIdeal(Adv).main : true ==> ={res}].
proof.
proc.
inline GIdeal(Adv).clientInitView GIdeal(Adv).clientGetView
  GIdeal(Adv).serverGenSec GIdeal(Adv).serverGetSec.
(* initialization, call to pre, sufficiency check *)
seq 5 6 :
  (={db, qrys, suff, glob BRO.BOr(RO.Or), glob Adv} /\
   G9.view{1} = GIdeal.clientView{2} /\ G9.view{1} = [] /\
   G9.sec{1} = GIdeal.servSec{2}).
swap {1} 1 4; swap {2} 1 5.
seq 4 5 :
  (={db, qrys, suff, glob BRO.BOr(RO.Or), glob Adv} /\
   G9.sec{1} = GIdeal.servSec{2}); first sim.
auto.
(* conditional *)
if => //.
(* then part *)
seq 1 1 :
  (={qrys, qrysCounts, glob BRO.BOr(RO.Or), glob Adv} /\
   G9.sec{1} = GIdeal.servSec{2} /\ G9.view{1} = GIdeal.clientView{2});
  first sim.
seq 1 1 :
  (={glob BRO.BOr(RO.Or), glob Adv} /\ G9.view{1} = GIdeal.clientView{2}).
call
  (_ :
   ={qrys, qrysCounts, glob BRO.BOr(RO.Or)} /\
   G9.view{1} = GIdeal.clientView{2} /\ G9.sec{1} = GIdeal.servSec{2} ==>
   ={glob BRO.BOr(RO.Or)} /\ G9.view{1} = GIdeal.clientView{2}).
proc; inline GIdeal(Adv).serverGetSec.
(* this is where we switch from BOr.hash to Or.hash: *)
sim; auto; progress; rewrite - appA; trivial.
auto. sim.
(* else part *)
sim; auto.
qed.

local lemma G9_GIdeal &m :
  Pr[G9.main() @ &m : res] = Pr[GIdeal(Adv).main() @ &m : res].
proof.
byequiv G9_GIdeal_main => //.
qed.

lemma GReal_GIdeal' &m :
  (forall (O <: BRO.BOR{Adv}),
   islossless O.bhash => islossless Adv(O).pre) =>
  (forall (O <: BRO.BOR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  `|Pr[GReal(Adv).main() @ &m : res] - Pr[GIdeal(Adv).main() @ &m : res]| <=
   (budget * (budget - 1))%r / (2 ^ tagLen)%r.
proof.
intros LossPre LossPost.
rewrite - (G9_GIdeal &m).
cut -> :
  (budget * (budget - 1))%r / (2 ^ tagLen)%r =
  2%r * (budget * (budget - 1))%r / (2 ^ (tagLen + 1))%r.
  cut -> : 2 ^ (tagLen + 1) = 2 * 2 ^ tagLen by smt.
  cut -> : (2 * 2 ^ tagLen)%r = 2%r * (2 ^ tagLen)%r by smt.
  cut -> :
    2%r * (budget * (budget - 1))%r / (2%r * (2 ^ tagLen)%r) =
    (2%r / 2%r) * ((budget * (budget - 1))%r / (2 ^ tagLen)%r) by smt.
  cut -> : 2%r / 2%r = 1%r by smt.
  smt.
apply (GReal_G9 &m) => //.
qed.

end section.

lemma GReal_GIdeal (Adv <: ADV{BRO.BOr(RO.Or), GReal, GIdeal}) &m :
  (forall (O <: BRO.BOR{Adv}),
   islossless O.bhash => islossless Adv(O).pre) =>
  (forall (O <: BRO.BOR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  `|Pr[GReal(Adv).main() @ &m : res] -
    Pr[GIdeal(Adv).main() @ &m : res]| <=
  (budget * (budget - 1))%r / (2 ^ tagLen)%r.
proof.
apply (GReal_GIdeal' Adv &m).
qed.
