(*******************************************************************************
  Copyright 2014-2015 MIT Lincoln Laboratory
  Project:     SPAR
  Author:      Alley Stoughton
  Description: Part of Private Count Retrieval (PCR) Protocol Proof
*******************************************************************************)

(* Protocol.ec *)

(* PCR Protocol and Related Definitions *)

(* This module defines the PCR protocol, and gives supporting
   defintions and lemmas. *)

prover "Alt-Ergo". timeout 5.  (* set prover and timeout *)

(***************************** Standard Theories ******************************)

require export Option.
require export Pair.
require export Int.
require export Real.
require export Distr.
require export FSet. export Dexcepted.
require export FMap.
require export List.
require Word.

(***************************** Auxiliary Theories *****************************)

require export OptionAux.
require export RealAux.
require export FMapAux.
require export ListAux.

(*********************** Secrets, Elements and Hash Tags **********************)

(* secret generated by the Server and shared with the Client, but not with
   Third Party (TP)

   a secret is a bitstring of length secLen *)

op secLen : int.  (* length of a secret *)

axiom SecLen : secLen >= 0.

clone Word as Sec with
  op length <- secLen
proof
  length_pos by smt.
(* the remaining axioms are unprovable specifications *)

type sec = Sec.word.  (* type of secrets *)

op zerosSec : sec = Sec.zeros.  (* the all 0 secret *)

(* uniform distribution on secrets *)

op secDistr : sec distr = Sec.Dword.dword.

lemma SecDistrChoice (sec : sec) :
  mu_x secDistr sec = 1%r / (2 ^ secLen)%r
by smt.

lemma SecDistrWeight : weight secDistr = 1%r
by smt.

(* type of elements *)

type elem.

op elemDefault : elem.  (* default element value *)

(* hash tags are bitstrings of length tagLen *)

op tagLen : int.

axiom TagLen : tagLen >= 0.

clone Word as Tag with
  op length <- tagLen
proof
  length_pos by smt.
(* the remaining axioms are unprovable specifications *)

type tag = Tag.word.  (* type of tags *)

op zerosTag : tag = Tag.zeros.  (* the all 0 tag *)

(* uniform distribution on tags *)

op tagDistr : tag distr = Tag.Dword.dword.

lemma TagDistrChoice (tag : tag) :
  mu_x tagDistr tag = 1%r / (2 ^ tagLen)%r
by smt.

lemma TagDistrWeight : weight tagDistr = 1%r
by smt.

(* if we subtract out less than all tags from tagDistr,
   the weight is still 1 *)

lemma TagDistrMinusWeight (tags : tag set) :
  card tags < 2 ^ tagLen =>
  weight (tagDistr \ tags) = 1%r
by smt.

(************************** Random Shuffling of Lists *************************)

require Shuffle.
clone Shuffle as Shuffle' with
  type t <- elem
proof *.
(* nothing to realize *)
export Shuffle'.

(******************************* Random Oracle ********************************)

require RandomOracles.  (* abstract theory *)

(* we clone RandomOracles as the theory RO, setting input to elem *
   sec, and fully realizing it *)

clone RandomOracles as RO with
  type input <- elem * sec,
  op inputDefault <- (elemDefault, zerosSec),
  op outputLen <- tagLen,
  type output <- tag,
  op outputDefault <- zerosTag,
  op outputDistr <- tagDistr
proof *.
(* beginning of realization *)
realize OutputLen. apply TagLen. qed.

realize OutputDistrChoice. apply TagDistrChoice. qed.

realize OutputDistrWeight. apply TagDistrWeight. qed.
(* end of realization *)

(****************** Databases, Queries and Hashed Databases *******************)

type db = elem list.  (* type of databases *)

type qrys = elem list.  (* type of lists of queries *)

type hdb = tag list.  (* type of hashed databases *)

(******************************* Parties' Views *******************************)

(* Server's view *)

type server_view_elem = [
    ServerViewGenSec   of sec           (* generated and shared secret *)
  | ServerViewDB       of db            (* received database *)
  | ServerViewShuffle  of db            (* shuffled database *)
  | ServerViewElemHash of (elem * tag)  (* hashed element, resulting in tag *)
  | ServerViewHDB      of hdb           (* shared hashed database *)
].

type server_view = server_view_elem list.

(* TP's view *)

type tp_view_elem = [
    TPViewHDB      of hdb          (* received hashed database *)
  | TPViewTagCount of (tag * int)  (* tag and tag count in hashed database *)
].

type tp_view = tp_view_elem list.

(* Client's view *)

type client_view_elem = [
    ClientViewQrys     of qrys                (* received list of queries *)
  | ClientViewGotSec   of sec                 (* received secret *)
  | ClientViewQryCount of (elem * tag * int)  (* query and its tag and count *)
].

type client_view = client_view_elem list.

(******************************** PCR Protocol ********************************)

(* module type of protocols

   initViews() must be run first, after which the xGetView functions may be
   called at any time

   after that, the following procedures should be run in this order:

     sec = serverGenSec()
     hdb = serverHashDB(db)
     tpAcceptHDB(hdb)
     clientLoop(qrys)
*)

module type PROTOCOL = {
  (* Views *)

  proc initViews() : unit              (* initialize all views *)
  proc serverGetView() : server_view   (* get Server's view *)
  proc tpGetView() : tp_view           (* get TP's view *)
  proc clientGetView() : client_view   (* get Client's view *)

  (* Server *)

  proc serverGenSec() : sec            (* make Server generate secret *)
  proc serverHashDB(db : db) : hdb     (* make Server hash database *)

  (* TP *)

  proc tpAcceptHDB(hdb : hdb) : unit   (* give TP the hashed database *)

  (* Client *)

  proc clientLoop(qrys : qrys) : unit  (* make Client process queries *)
}.

(* the PCR protocol *)

module Protocol : PROTOCOL = {
  module Or = RO.Or  (* random oracle *)

  (* Views *)

  var serverView : server_view  (* Server's view *)
  var tpView : tp_view          (* TP's view *)
  var clientView : client_view  (* Client's view *)

  proc initViews() : unit = {
    serverView = [];
    tpView = [];
    clientView = [];
  }

  proc serverGetView() : server_view = {
    return serverView;
  }

  proc tpGetView() : tp_view = {
    return tpView;
  }

  proc clientGetView() : client_view = {
    return clientView;
  }

  (* Server *)

  var serverSec : sec  (* Server's secret *)

  proc serverGenSec() : sec = {  (* must be called first, only once *)
    serverSec = $secDistr;
    serverView = serverView ++ [ServerViewGenSec serverSec];
    return serverSec;
  }

  proc serverGetSec() : sec = {  (* called by clientLoop *)
    return serverSec;
  }

  proc serverHashDB(db : db) : hdb = {
    var i : int;
    var elem : elem;
    var tag : tag;
    var hdb : tag list;
    serverView = serverView ++ [ServerViewDB db];
    db = Shuffle.shuffle(db);
    serverView = serverView ++ [ServerViewShuffle db];
    hdb = [];
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      tag = Or.hash((elem, serverSec));
      hdb = hdb ++ [tag];
      serverView = serverView ++ [ServerViewElemHash(elem, tag)];
      i = i + 1;
    }
    serverView = serverView ++ [ServerViewHDB hdb];
    return hdb;
  }

  (* TP *)

  var tpHDB : hdb  (* hashed database *)

  proc tpAcceptHDB(hdb : hdb) : unit = {  (* must be called first, only once *)
    tpHDB = hdb;
    tpView = tpView ++ [TPViewHDB hdb];
  }

  proc tpCountTag(tag : tag) : int = {  (* called by clientLoop *)
    var i, n : int;
    n = 0;
    i = 0;
    while (i < length tpHDB) {
      if (oget(nth tpHDB i) = tag) {
        n = n + 1;
      }
      i = i + 1;
    }
    tpView = tpView ++ [TPViewTagCount(tag, n)];
    return n;
  }

  (* Client *)

  proc clientLoop(qrys : qrys) : unit = {
    var i, n : int;
    var tag : tag;
    var qry : elem;
    var sec : sec;
    clientView = clientView ++ [ClientViewQrys qrys];
    sec = serverGetSec();
    clientView = clientView ++ [ClientViewGotSec sec];
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = Or.hash((qry, sec));
      n = tpCountTag(tag);
      clientView = clientView ++ [ClientViewQryCount(qry, tag, n)];
      i = i + 1;
    }
  }
}.

(*********************************** Games ************************************)

(* a game is a module that provides a main procedure taking in no arguments
   and returning a boolean *)

module type GAME = {
  proc main() : bool
}.
