(*******************************************************************************
  Copyright 2014-2015 MIT Lincoln Laboratory
  Project:     SPAR
  Author:      Alley Stoughton
  Description: Part of Private Count Retrieval (PCR) Protocol Proof
*******************************************************************************)

(* Server.ec *)

(* Proof of Security Against Server *)

prover "Alt-Ergo". timeout 10.  (* set prover and timeout *)

(******************** PCR Protocol and Related Definitions ********************)

require import Protocol.

(********************************* Adversary **********************************)

(* Adversary module type, parameterized by a random oracle

   pre takes in the secret, and produces a database and list of
   queries, possibly doing hashing in the process

   post takes in a view and returns a boolean, possibly doing hashing
   in the process *)

module type ADV(O : RO.OR) = {
  proc * pre(sec : sec) : db * qrys {O.hash}
  proc post(view : server_view) : bool {O.hash}
}.

(*************************** Real and Ideal Games *****************************)

(* the "real" game *)

module GReal(Adv : ADV) : GAME = {
  module Or = RO.Or    (* random oracle *)
  module A  = Adv(Or)  (* unlimited access to random oracle *)

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b : bool;
    var sec : sec;
    var hdb : hdb;
    var view : server_view;
    Protocol.initViews();
    Or.init();
    sec = Protocol.serverGenSec();
    (db, qrys) = A.pre(sec);
    hdb = Protocol.serverHashDB(db);
    Protocol.tpAcceptHDB(hdb);
    Protocol.clientLoop(qrys);
    view = Protocol.serverGetView();
    b = A.post(view);
    return b;
  }
}.

(* the "ideal" game

   Client/TP interaction removed *)

module GIdeal(Adv : ADV) : GAME = {
  module Or = RO.Or  (* random oracle *)

  (* View *)

  var serverView : server_view  (* Server's view *)

  proc serverInitView() : unit = {  (* must be called first, only once *)
    serverView = [];
  }

  proc serverGetView() : server_view = {
    return serverView;
  }

  (* Server *)

  var serverSec : sec  (* Server's secret *)

  proc serverGenSec() : sec = {
    serverSec = $secDistr;
    serverView = serverView ++ [ServerViewGenSec serverSec];
    return serverSec;
  }

  proc serverHashDB(db : db) : hdb = {
    var i : int;
    var elem : elem;
    var tag : tag;
    var hdb : tag list;
    serverView = serverView ++ [ServerViewDB db];
    db = Shuffle.shuffle(db);
    serverView = serverView ++ [ServerViewShuffle db];
    hdb = [];
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      tag = Or.hash((elem, serverSec));
      hdb = hdb ++ [tag];
      serverView = serverView ++ [ServerViewElemHash(elem, tag)];
      i = i + 1;
    }
    serverView = serverView ++ [ServerViewHDB hdb];
    return hdb;
  }

  (* Main Processing *)

  module A = Adv(Or)

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b : bool;
    var sec : sec;
    var hdb : hdb;
    var view : server_view;
    serverInitView();
    Or.init();
    sec = serverGenSec();
    (db, qrys) = A.pre(sec);  (* qrys never used, below *)
    hdb = serverHashDB(db);  (* hdb never used, below *)
    view = serverGetView();
    b = A.post(view);
    return b;
  }
}.

(* see end-of-file for top-level theorem *)

(********************************* Proof Body *********************************)

(************************* Theories Supporting Proof **************************)

(* disjoint unions *)

require import Union.

(******************************** Proof Section *******************************)

(* the rest of the proof is within a section, in which the Adversary,
   Adv, is declared locally *)

section.

(* declare Adversary module -- subsequent games will reference it, instead
   of being parameterized by it *)

declare module Adv : ADV{RO.Or, GReal, GIdeal}.

(* G1 is the result of doing inlining and dead code elimination to
   GReal; the hashing of ClientLoop is put in the procedure
   redudantHashing *)

local module G1 : GAME = {
  var view : server_view
  var sec : sec

  proc serverHashDB(db : db) : hdb = {
    var i : int;
    var elem : elem;
    var tag : tag;
    var hdb : tag list;
    view = view ++ [ServerViewDB db];
    db = Shuffle.shuffle(db);
    view = view ++ [ServerViewShuffle db];
    hdb = [];
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      tag = RO.Or.hash((elem, sec));
      hdb = hdb ++ [tag];
      view = view ++ [ServerViewElemHash(elem, tag)];
      i = i + 1;
    }
    view = view ++ [ServerViewHDB hdb];
    return hdb;
  }

  proc redundantHashing(qrys : qrys) : unit = {
    var i, n : int;
    var tag : tag;
    var qry : elem;
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = RO.Or.hash((qry, sec));
      i = i + 1;
    }
  }

  module A = Adv(RO.Or)

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b : bool;
    var hdb : hdb;
    RO.Or.init();
    sec = $secDistr;
    view = [ServerViewGenSec sec];
    (db, qrys) = A.pre(sec);
    hdb = serverHashDB(db);
    redundantHashing(qrys);
    b = A.post(view);
    return b;
  }
}.

(* G2 is the result of doing inlining and dead code elimination to
   GIdeal *)

local module G2 : GAME = {
  var view : server_view
  var sec : sec

  proc serverHashDB(db : db) : hdb = {
    var i : int;
    var elem : elem;
    var tag : tag;
    var hdb : tag list;
    view = view ++ [ServerViewDB db];
    db = Shuffle.shuffle(db);
    view = view ++ [ServerViewShuffle db];
    hdb = [];
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      tag = RO.Or.hash((elem, sec));
      hdb = hdb ++ [tag];
      view = view ++ [ServerViewElemHash(elem, tag)];
      i = i + 1;
    }
    view = view ++ [ServerViewHDB hdb];
    return hdb;
  }

  module A = Adv(RO.Or)

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b : bool;
    var hdb : hdb;
    RO.Or.init();
    sec = $secDistr;
    view = [ServerViewGenSec sec];
    (db, qrys) = A.pre(sec);
    hdb = serverHashDB(db);
    (* no call to redundantHashing *)
    b = A.post(view);
    return b;
  }
}.

local lemma Protocol_G1_serverHashDB :
  equiv
  [Protocol.serverHashDB ~ G1.serverHashDB :
   ={db, RO.Or.mp} /\ Protocol.serverSec{1} = G1.sec{2} /\
   Protocol.serverView{1} = G1.view{2} ==>
   ={res, RO.Or.mp} /\ Protocol.serverSec{1} = G1.sec{2} /\
   Protocol.serverView{1} = G1.view{2}].
proof.
sim.
qed.

local lemma GReal_G1_main :
  equiv
  [GReal(Adv).main ~ G1.main :
   true ==> ={res}].
proof.
proc.
inline Protocol.initViews Protocol.serverGenSec Protocol.serverGetView
  Protocol.tpAcceptHDB.
(* handle initialization *)
seq 7 3 :
  (={RO.Or.mp} /\ Protocol.serverSec{1} = G1.sec{2} /\
   Protocol.serverView{1} = G1.view{2} /\ sec{1} = Protocol.serverSec{1}).
inline RO.Or.init; auto.
(* handle call to pre *)
seq 1 1 :
  (={RO.Or.mp, glob Adv, db, qrys} /\ Protocol.serverSec{1} = G1.sec{2} /\
   Protocol.serverView{1} = G1.view{2} /\ sec{1} = G1.sec{2}).
sim.
(* handle call to serverHashDB *)
seq 1 1 :
  (={RO.Or.mp, glob Adv, db, qrys} /\ Protocol.serverSec{1} = G1.sec{2} /\
   Protocol.serverView{1} = G1.view{2}).
call Protocol_G1_serverHashDB => //.
(* skip hdb assignments in {1} *)
seq 3 0 :
  (={RO.Or.mp, glob Adv, db, qrys} /\ Protocol.serverSec{1} = G1.sec{2} /\
   Protocol.serverView{1} = G1.view{2}).
auto.
(* handle redundant hashing *)
seq 2 1 :
  (={RO.Or.mp, glob Adv, db} /\ Protocol.serverSec{1} = G1.sec{2} /\
   view{1} = G1.view{2}).
inline Protocol.clientLoop G1.redundantHashing.
inline Protocol.serverGetSec.
seq 5 2 :
  (={RO.Or.mp, glob Adv, db} /\ Protocol.serverSec{1} = G1.sec{2} /\
   sec0{1} = G1.sec{2} /\ Protocol.serverView{1} = G1.view{2} /\
   ={i, qrys0} /\ i{1} = 0).
auto. wp.
while (={i, qrys0, RO.Or.mp} /\ sec0{1} = G1.sec{2}).
seq 2 2 : (={i, qrys0, RO.Or.mp, qry, tag} /\ sec0{1} = G1.sec{2}).
call (_ : ={RO.Or.mp, inp} ==> ={RO.Or.mp, res}); first sim.
auto.
inline Protocol.tpCountTag.
wp; sp 3 0.
while {1}
  (i0{1} <= length Protocol.tpHDB{1}) (length Protocol.tpHDB{1} - i0{1}).
intros &m z; auto; smt.
skip; smt.
skip; smt.
(* handle call to post *)
sim.
qed.

local lemma GReal_G1 &m :
  Pr[GReal(Adv).main() @ &m : res] = Pr[G1.main() @ &m : res].
proof.
byequiv GReal_G1_main => //.
qed.

(* we show that G1 and G2 are equivalent using the RedundantHashing
   abstract theory of RO

   we locally clone the abstract theory RO.RedundantHashing as the
   theory RH, fully realizing it (there is nothing to realize) *)

local clone RO.RedundantHashing as RH
proof *.

(* we begin by defining a redundant hashing adversary RHAdv plus a
   hashing module Hashing, so that RH.GDoHashing(RHAdv, Hashing) is
   equivalent to G1, and RH.GNoHashing(RHAdv) is equivalent to G2 *)

local module RHAdv(O : RO.OR) : RH.RHADV(O) = {
  var view : server_view
  var sec : sec

  proc serverHashDB(db : db) : hdb = {
    var i : int;
    var elem : elem;
    var tag : tag;
    var hdb : tag list;
    view = view ++ [ServerViewDB db];
    db = Shuffle.shuffle(db);
    view = view ++ [ServerViewShuffle db];
    hdb = [];
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      tag = O.hash((elem, sec));
      hdb = hdb ++ [tag];
      view = view ++ [ServerViewElemHash(elem, tag)];
      i = i + 1;
    }
    view = view ++ [ServerViewHDB hdb];
    return hdb;
  }

  module A = Adv(O)

  proc pre() : RH.rh_pre_res = {
    var db : db;
    var qrys : qrys;
    var b : bool;
    var hdb : hdb;
    var rhpr : RH.rh_pre_res;
    (* no oracle initialization *)
    sec = $secDistr;
    view = [ServerViewGenSec sec];
    (db, qrys) = A.pre(sec);
    hdb = serverHashDB(db);
    (* no call to redundantHashing *)
    rhpr = Right(map (fun elem, (elem, sec)) qrys);
    return rhpr;
  }

  proc post() : bool = {
    var b : bool;
    b = A.post(view);
    return b;
  }
}.

local module Hashing(O : RO.OR) : RH.HASHING(O) = {
  proc hashing(inps : (elem * sec)list) : unit = {
    var i : int;
    var inp : elem * sec;
    var tag : tag;
    i = 0;
    while (i < length inps) {
      inp = oget(nth inps i);
      tag = O.hash(inp);
      i = i + 1;
    }
  }
}.

local lemma Hashing_Lossless_hashing :
  forall (O <: RO.OR{Hashing}),
  islossless O.hash => islossless Hashing(O).hashing.
proof.
intros O LossHash.
proc.
while (true) (length inps - i).
auto.
call (_ : true ==> true); first apply LossHash.
auto; smt.
auto; smt.
qed.

local lemma G1_GDoHashing_RHAdv_Hashing_main :
  equiv[G1.main ~ RH.GDoHashing(RHAdv, Hashing).main : true ==> ={res}].
proc.
inline
  RH.GDoHashing(RHAdv, Hashing).RHA.pre
  RH.GDoHashing(RHAdv, Hashing).RHA.post.
(* initialization, calls to pre and serverHashDB *)
seq 5 5 :
  (={glob RO.Or, glob Adv, db, qrys} /\ G1.sec{1} = RHAdv.sec{2} /\
  G1.view{1} = RHAdv.view{2}).
sim.
(* simplification of {2} *)
sp 0 2; rcondf {2} 1; auto.
(* redundant hashing *)
seq 0 1 :
  (={glob RO.Or, glob Adv, qrys} /\ G1.view{1} = RHAdv.view{2} /\
   inps{2} = map (fun (elem : elem), (elem, G1.sec{1})) qrys{2});
  first auto.
seq 1 1 : (={glob RO.Or, glob Adv} /\ G1.view{1} = RHAdv.view{2}).
inline G1.redundantHashing RH.GDoHashing(RHAdv, Hashing).H.hashing.
seq 2 2 :
  (={glob RO.Or, glob Adv, i} /\ G1.view{1} = RHAdv.view{2} /\
  0 <= i{1} <= length qrys0{1} /\
  inps0{2} = map (fun (elem : elem), (elem, G1.sec{1})) qrys0{1} /\
  length qrys0{1} = length inps0{2}).
auto; smt.
while
  (={glob RO.Or, glob Adv, i} /\ G1.view{1} = RHAdv.view{2} /\
   0 <= i{1} <= length qrys0{1} /\
   inps0{2} = map (fun (elem : elem), (elem, G1.sec{1})) qrys0{1} /\
   length qrys0{1} = length inps0{2}).
wp.
call (_ : ={glob RO.Or, inp} ==> ={glob RO.Or, res}); first sim.
auto; smt.
auto; smt.
(* call to post *)
sim.
qed.

local lemma G1_GDoHashing_RHAdv_Hashing &m :
  Pr[G1.main() @ &m : res] =
  Pr[RH.GDoHashing(RHAdv, Hashing).main() @ &m : res].
proof.
byequiv G1_GDoHashing_RHAdv_Hashing_main => //.
qed.

local lemma GNoHashing_RHAdv_G2_main :
  equiv[RH.GNoHashing(RHAdv).main ~ G2.main : true ==> ={res}].
proc.
inline RH.GNoHashing(RHAdv).RHA.pre RH.GNoHashing(RHAdv).RHA.post.
(* initialization, calls to pre and serverHashDB *)
seq 5 5 :
  (={glob RO.Or, glob Adv, db, qrys} /\ RHAdv.sec{1} = G2.sec{2} /\
  RHAdv.view{1} = G2.view{2}).
sim.
(* simplification of {1} *)
sp 2 0; rcondf {1} 1; auto.
(* call to post *)
sim.
qed.

local lemma GNoHashing_RHAdv_G2 &m :
  Pr[RH.GNoHashing(RHAdv).main() @ &m : res] =
  Pr[G2.main() @ &m : res].
proof.
byequiv GNoHashing_RHAdv_G2_main => //.
qed.

local lemma G1_G2 &m :
  Pr[G1.main() @ &m : res] = Pr[G2.main() @ &m : res].
proof.
rewrite (G1_GDoHashing_RHAdv_Hashing &m).
rewrite - (GNoHashing_RHAdv_G2 &m).
apply (RH.GDoHashing_GNoHashing RHAdv Hashing &m).
apply Hashing_Lossless_hashing.
qed.

local lemma GReal_G2 &m :
  Pr[GReal(Adv).main() @ &m : res] = Pr[G2.main() @ &m : res].
proof.
rewrite (GReal_G1 &m).
apply (G1_G2 &m).
qed.

local lemma G2_GIdeal_serverHashDB :
  equiv
  [G2.serverHashDB ~ GIdeal(Adv).serverHashDB :
   ={db, RO.Or.mp} /\ G2.sec{1} = GIdeal.serverSec{2} /\
   G2.view{1} = GIdeal.serverView{2} ==>
   ={res, RO.Or.mp} /\ G2.sec{1} = GIdeal.serverSec{2} /\
   G2.view{1} = GIdeal.serverView{2}].
proof.
sim.
qed.

local lemma G2_GIdeal_main :
  equiv
  [G2.main ~ GIdeal(Adv).main :
   true ==> ={res}].
proof.
proc.
inline GIdeal(Adv).serverInitView GIdeal(Adv).serverGenSec
  GIdeal(Adv).serverGetView.
(* handle initialization *)
seq 3 5 :
  (={RO.Or.mp} /\ G2.sec{1} = GIdeal.serverSec{2} /\
   G2.sec{1} = sec{2} /\ G2.view{1} = GIdeal.serverView{2}).
inline RO.Or.init; auto.
(* handle call to pre *)
seq 1 1 :
  (={RO.Or.mp, glob Adv, db, qrys} /\ G2.sec{1} = GIdeal.serverSec{2} /\
   G2.sec{1} = sec{2} /\ G2.view{1} = GIdeal.serverView{2}).
sim.
(* handle call to serverHashDB *)
seq 1 1 :
  (={RO.Or.mp, glob Adv, db, qrys} /\ G2.sec{1} = GIdeal.serverSec{2} /\
   G2.view{1} = GIdeal.serverView{2}).
call G2_GIdeal_serverHashDB => //.
(* handle finalization and call to post *)
sim.
qed.

local lemma G2_GIdeal &m :
  Pr[G2.main() @ &m : res] = Pr[GIdeal(Adv).main() @ &m : res].
proof.
byequiv G2_GIdeal_main => //.
qed.

lemma GReal_GIdeal' &m :
  Pr[GReal(Adv).main() @ &m : res] = Pr[GIdeal(Adv).main() @ &m : res].
proof.
rewrite (GReal_G2 &m).
apply (G2_GIdeal &m).
qed.

end section.

lemma GReal_GIdeal (Adv <: ADV{RO.Or, GReal, GIdeal}) &m :
  Pr[GReal(Adv).main() @ &m : res] = Pr[GIdeal(Adv).main() @ &m : res].
proof.
apply (GReal_GIdeal' Adv &m).
qed.
