(*******************************************************************************
  Copyright 2014-2015 MIT Lincoln Laboratory
  Project:     SPAR
  Author:      Alley Stoughton
  Description: Part of Private Count Retrieval (PCR) Protocol Proof
*******************************************************************************)

(* TP.ec *)

(* Proof of Security Against TP (Third Party) *)

prover "Alt-Ergo". timeout 10.  (* set prover and timeout *)

(******************** PCR Protocol and Related Definitions ********************)

require import Protocol.

(********************* Limited and Private Random Oracles *********************)

(* hashing limit *)

op limit : int.

axiom Limit : limit >= 0.

(* limited random oracle *)

clone RO.Limited as ROLim with
  op limit <- limit
proof *.
(* beginning of realization *)
realize Limit. apply Limit. qed.
(* end of realization *)

(* private random oracle for hashing elements (not element/secret pairs) *)

clone RandomOracles as Priv with
  type input <- elem,
  op inputDefault <- elemDefault,
  op outputLen <- tagLen,
  type output <- tag,
  op outputDefault <- zerosTag,
  op outputDistr <- tagDistr
proof *.
(* beginning of realization *)
realize OutputLen. apply TagLen. qed.

realize OutputDistrChoice. apply TagDistrChoice. qed.

realize OutputDistrWeight. apply TagDistrWeight. qed.
(* end of realization *)

(********************************* Adversary **********************************)

(* Adversary module type, parameterized by a random oracle

   Adversary *doesn't* have access to the secret

   pre produces a database and list of queries, possibly doing hashing
   in the process

   post takes in a view and returns a boolean, possibly doing hashing
   in the process *)

module type ADV(O : RO.OR) = {
  proc * pre() : db * qrys {O.hash}
  proc post(view : tp_view) : bool {O.hash}
}.

(*************************** Real and Ideal Games *****************************)

(* the "real" game

   secret is generated by Server and shared with Client

   Adversary has limited access to random oracle

   Server and Client hash elements paired with this secret, using
   (unlimited) random oracle *)

module GReal(Adv : ADV) : GAME = {
  module Or  = RO.Or          (* random oracle *)
  module LOr = ROLim.LOr(Or)  (* limited random oracle built from Or *)

  module A = Adv(LOr)

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b : bool;
    var sec : sec;
    var hdb : hdb;
    var view : tp_view;
    Protocol.initViews();
    LOr.init();  (* also initializes Or *)
    sec = Protocol.serverGenSec();  (* sec unused, below *)
    (db, qrys) = A.pre();
    hdb = Protocol.serverHashDB(db);
    Protocol.tpAcceptHDB(hdb);
    Protocol.clientLoop(qrys);
    view = Protocol.tpGetView();
    b = A.post(view);
    return b;
  }
}.

(* the "ideal" game

   doesn't make use of Server/Client shared secret

   Adversary has limited access to random oracle

   Server and Client hash elements using a private random oracle *)

module GIdeal(Adv : ADV) : GAME = {
  module Or     = RO.Or          (* random oracle *)
  module LOr    = ROLim.LOr(Or)  (* limited random oracle built from Or *)
  module OrPriv = Priv.Or        (* private random oracle for elements *)

  (* View *)

  var tpView : tp_view  (* TP's view *)

  proc tpInitView() : unit = {  (* must be called first, only once *)
    tpView = [];
  }

  proc tpGetView() : tp_view = {
    return tpView;
  }

  (* Server *)

  proc serverHashDB(db : db) : hdb = {
    var i : int;
    var elem : elem;
    var tag : tag;
    var hdb : tag list;
    db = Shuffle.shuffle(db);
    hdb = [];
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      tag = OrPriv.hash(elem);  (* note: private oracle *)
      hdb = hdb ++ [tag];
      i = i + 1;
    }
    return hdb;
  }

  (* TP *)

  var tpHDB : hdb  (* hashed database *)

  proc tpAcceptHDB(hdb : hdb) : unit = {
    tpHDB = hdb;
    tpView = tpView ++ [TPViewHDB hdb];
  }

  proc tpCountTag(tag : tag) : int = {
    var i, n : int;
    n = 0;
    i = 0;
    while (i < length tpHDB) {
      if (oget(nth tpHDB i) = tag) {
        n = n + 1;
      }
      i = i + 1;
    }
    tpView = tpView ++ [TPViewTagCount(tag, n)];
    return n;
  }

  (* Client *)

  proc clientLoop(qrys : qrys) : unit = {
    var i, n : int;
    var tag : tag;
    var qry : elem;
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = OrPriv.hash(qry);  (* note: private oracle *)
      n = tpCountTag(tag);
      i = i + 1;
    }
  }

  (* Main Processing *)

  module A = Adv(LOr)

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b : bool;
    var hdb : hdb;
    var view : tp_view;
    tpInitView();
    LOr.init();  (* also initializes Or *)
    OrPriv.init();
    (db, qrys) = A.pre();
    hdb = serverHashDB(db);
    tpAcceptHDB(hdb);
    clientLoop(qrys);
    view = tpGetView();
    b = A.post(view);
    return b;
  }
}.

(* see end-of-file for top-level theorem *)

(********************************* Proof Body *********************************)

(************************* Theories Supporting Proof **************************)

(* secrecy random oracles *)

require SecrecyRandomOracle.  (* abstract theory *)

(******************************** Proof Section *******************************)

(* the rest of the proof is within a section, in which the Adversary,
   Adv, is declared locally *)

section.

declare module Adv : ADV{ROLim.LOr(RO.Or), Priv.Or, GReal, GIdeal}.

(* G1 is like GReal, except that is uses Adv, instead of being
   parameterized by an Adversary, inlining and dead code elimination
   have been done, and the Server and Client do their hashing using a
   new hashElem procedure that takes in an element and hashes that
   element paired with the shared secret *)

local module G1 : GAME = {
  var view : tp_view
  var sec : sec
  var tpHDB : hdb

  proc hashElem(elem : elem) : tag = {
    var tag : tag;
    tag = RO.Or.hash((elem, sec));
    return tag;
  }

  proc serverHashDB(db : db) : hdb = {
    var i : int;
    var elem : elem;
    var tag : tag;
    var hdb : tag list;
    db = Shuffle.shuffle(db);
    hdb = [];
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      tag = hashElem(elem);
      hdb = hdb ++ [tag];
      i = i + 1;
    }
    return hdb;
  }

  proc tpCountTag(tag : tag) : int = {
    var i, n : int;
    n = 0;
    i = 0;
    while (i < length tpHDB) {
      if (oget(nth tpHDB i) = tag) {
        n = n + 1;
      }
      i = i + 1;
    }
    view = view ++ [TPViewTagCount(tag, n)];
    return n;
  }

  proc clientLoop(qrys : qrys) : unit = {
    var i, n : int;
    var tag : tag;
    var qry : elem;
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = hashElem(qry);
      n = tpCountTag(tag);
      i = i + 1;
    }
  }

  module A = Adv(ROLim.LOr(RO.Or))

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b : bool;
    ROLim.LOr(RO.Or).init();
    sec = $secDistr;
    (db, qrys) = A.pre();
    tpHDB = serverHashDB(db);
    view = [TPViewHDB tpHDB];
    clientLoop(qrys);
    b = A.post(view);
    return b;
  }
}.

local lemma Protocol_G1_serverHashDB :
  equiv
  [Protocol.serverHashDB ~ G1.serverHashDB :
   ={db, glob RO.Or} /\ Protocol.serverSec{1} = G1.sec{2} ==>
   ={res, glob RO.Or}].
proof.
proc.
inline G1.hashElem.
sim.
qed.

local lemma Protocol_G1_clientLoop :
  equiv
  [Protocol.clientLoop ~ G1.clientLoop :
   ={qrys, glob RO.Or} /\ Protocol.serverSec{1} = G1.sec{2} /\
   Protocol.tpHDB{1} = G1.tpHDB{2} /\ Protocol.tpView{1} = G1.view{2} ==>
   ={glob RO.Or} /\ Protocol.tpHDB{1} = G1.tpHDB{2} /\
   Protocol.tpView{1} = G1.view{2}].
proof.
proc.
inline Protocol.serverGetSec G1.hashElem.
sim.
qed.

local lemma GReal_G1_main :
  equiv
  [GReal(Adv).main ~ G1.main :
   true ==> ={res}].
proof.
proc.
inline Protocol.initViews Protocol.serverGenSec
  Protocol.tpAcceptHDB Protocol.tpGetView.
(* handle oracle initialization and secret generation *)
seq 7 2 :
  (={glob ROLim.LOr(RO.Or)} /\ Protocol.serverSec{1} = G1.sec{2} /\
   Protocol.tpView{1} = []).
wp; sp; rnd.
inline ROLim.LOr(RO.Or).init RO.Or.init; auto.
(* handle call to pre *)
seq 1 1 :
  (={glob ROLim.LOr(RO.Or), glob Adv, db, qrys} /\
   Protocol.serverSec{1} = G1.sec{2} /\ Protocol.tpView{1} = []).
call
  (_ :
   ={glob ROLim.LOr(RO.Or)} ==>
   ={res, glob ROLim.LOr(RO.Or), glob Adv}).
sim.
auto.
(* handle call to serverHashDB and initialiation of tpHDB and view *)
seq 4 2 :
  (={glob ROLim.LOr(RO.Or), glob Adv, db, qrys} /\
   Protocol.serverSec{1} = G1.sec{2} /\ Protocol.tpHDB{1} = G1.tpHDB{2} /\
   Protocol.tpView{1} = G1.view{2}).
wp.
call
  (_ :
   ={db, glob RO.Or} /\ Protocol.serverSec{1} = G1.sec{2} ==>
   ={res, glob RO.Or}).
apply Protocol_G1_serverHashDB.
auto.
(* handle call to clientLoop and initialization of view{1} *)
seq 2 1 :
  (={glob ROLim.LOr(RO.Or), glob Adv, db, qrys} /\
   Protocol.serverSec{1} = G1.sec{2} /\ Protocol.tpHDB{1} = G1.tpHDB{2} /\
   Protocol.tpView{1} = G1.view{2} /\ view{1} = Protocol.tpView{1}).
wp.
call
  (_ :
   ={qrys, glob RO.Or} /\ Protocol.serverSec{1} = G1.sec{2} /\
   Protocol.tpHDB{1} = G1.tpHDB{2} /\ Protocol.tpView{1} = G1.view{2} ==>
   ={glob RO.Or} /\ Protocol.tpHDB{1} = G1.tpHDB{2} /\
   Protocol.tpView{1} = G1.view{2}).
apply Protocol_G1_clientLoop.
auto.
(* handle call to post *)
sim.
qed.

local lemma GReal_G1 &m :
  Pr[GReal(Adv).main() @ &m : res] = Pr[G1.main() @ &m : res].
proof.
byequiv GReal_G1_main => //.
qed.

(* we locally clone the abstract theory SecrecyRandomOracle as the
   theory SRO, fully realizing it *)

local clone SecrecyRandomOracle as SRO with
  type elem <- elem,
  op outputLen <- tagLen,
  type output <- tag,
  op outputDefault <- zerosTag,
  op outputDistr <- tagDistr,
  op secLen <- secLen,
  type sec <- sec,
  op secDefault <- zerosSec,
  op secDistr <- secDistr,
  op limit <- limit
proof *.
(* beginning of realization *)
realize OutputLen. apply TagLen. qed.

realize OutputDistrChoice. apply TagDistrChoice. qed.

realize OutputDistrWeight. apply TagDistrWeight. qed.

realize SecLen. apply SecLen. qed.

realize SecDistrChoice. apply SecDistrChoice. qed.

realize SecDistrWeight. apply SecDistrWeight. qed.

realize Limit. apply Limit. qed.
(* end of realization *)

(* convert a secrecy random oracle to a random oracle (to be given to
   adversary, which lacks access to init) *)

local module Conv(SOr : SRO.SEC_OR) : RO.OR = {
  proc init() : unit = {  (* dummy -- just for typechecking *)
  }

  proc hash(inp : elem * sec) : tag = {
    var tag : tag;
    tag = SOr.lhash(inp);  (* limited hashing *)
    return tag;
  }
}.

(* concrete secrecy adversary, parameterized by a secrecy random
   oracle

   like G1, except that the use of the shared secret has been removed,
   oracle initialization has been removed, calls to hashElem have been
   replaced by calls to the hash procedure of the secrecy random
   oracle, and Adv is given the random oracle derived by Conv from the
   secrecy random oracle *)

local module SecAdv(SOr : SRO.SEC_OR) : SRO.SEC_ADV(SOr) = {
  var view : tp_view
  var tpHDB : hdb

  proc serverHashDB(db : db) : hdb = {
    var i : int;
    var elem : elem;
    var tag : tag;
    var hdb : tag list;
    db = Shuffle.shuffle(db);
    hdb = [];
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      tag = SOr.hash(elem);
      hdb = hdb ++ [tag];
      i = i + 1;
    }
    return hdb;
  }

  proc tpCountTag(tag : tag) : int = {
    var i, n : int;
    n = 0;
    i = 0;
    while (i < length tpHDB) {
      if (oget(nth tpHDB i) = tag) {
        n = n + 1;
      }
      i = i + 1;
    }
    view = view ++ [TPViewTagCount(tag, n)];
    return n;
  }

  proc clientLoop(qrys : qrys) : unit = {
    var i, n : int;
    var tag : tag;
    var qry : elem;
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = SOr.hash(qry);
      n = tpCountTag(tag);
      i = i + 1;
    }
  }

  module A = Adv(Conv(SOr))

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b : bool;
    var hdb : hdb;
    (db, qrys) = A.pre();
    tpHDB = serverHashDB(db);
    view = [TPViewHDB tpHDB];
    clientLoop(qrys);
    b = A.post(view);
    return b;
  }
}.

(* we need to show that the main procedure of SecAdv is lossless *)

local lemma SecAdv_Lossless_serverHashDB :
  forall (SOr <: SRO.SEC_OR),
  islossless SOr.hash =>
  islossless SecAdv(SOr).serverHashDB.
proof.
intros SOr LossHash.
proc.
while (true) (length db - i).
auto.
call (_ : true ==> true); first apply LossHash.
auto; smt.
wp.
call (_ : true ==> true); first apply Lossless_shuffle.
auto; smt.
qed.

local lemma SecAdv_Lossless_tpCountTag :
  forall (SOr <: SRO.SEC_OR),
  islossless SecAdv(SOr).tpCountTag.
proof.
intros SOr.
proc.
wp.
while (true) (length SecAdv.tpHDB - i).
intros z.
auto; smt.
auto; smt.
qed.

local lemma SecAdv_Lossless_clientLoop :
  forall (SOr <: SRO.SEC_OR),
  islossless SOr.hash =>
  islossless SecAdv(SOr).clientLoop.
proof.
intros SOr LossHash.
proc.
while (true) (length qrys - i).
auto.
call (_ : true ==> true).
apply (SecAdv_Lossless_tpCountTag SOr).
call (_ : true ==> true); first apply LossHash.
auto; smt.
auto; smt.
qed.

local lemma SecAdv_Lossless_main :
  (forall (Or <: RO.OR{Adv}),
   islossless Or.hash => islossless Adv(Or).pre) =>
  (forall (Or <: RO.OR{Adv}),
   islossless Or.hash => islossless Adv(Or).post) =>
  (forall (SOr <: SRO.SEC_OR{SecAdv}),
   islossless SOr.lhash => islossless SOr.hash =>
   islossless SecAdv(SOr).main).
proof.
intros LossPre LossPost SOr LossBHash LossHash.
proc.
call (_ : true ==> true).
apply (LossPost (Conv(SOr))).
proc.
call (_ : true ==> true).
apply LossBHash.
auto.
call (_ : true ==> true).
apply (SecAdv_Lossless_clientLoop SOr).
apply LossHash.
wp.
call (_ : true ==> true).
apply (SecAdv_Lossless_serverHashDB SOr).
apply LossHash.
call (_ : true ==> true).
apply (LossPre (Conv(SOr))).
proc.
call (_ : true ==> true).
apply LossBHash.
auto.
auto.
qed.

(* preparation for showing equivalence between G1 and SRO.GDep(SecAdv) *)

local lemma G1_SRO_SecOrDep_hashElem_hash (xs : (elem * sec)set) :
  equiv
  [G1.hashElem ~ SRO.SecOrDep.hash :
   ={elem} /\
   G1.sec{1} = SRO.SecOrDep.secret{2} /\
   RO.Or.mp{1} = SRO.SecOrDep.mp{2} /\ xs = dom RO.Or.mp{1} ==>
   ={res} /\
   RO.Or.mp{1} = SRO.SecOrDep.mp{2} /\ xs <= dom RO.Or.mp{1}].
proof.
proc.
inline RO.Or.hash.
seq 1 0 :
  (={elem} /\ G1.sec{1} = SRO.SecOrDep.secret{2} /\
   RO.Or.mp{1} = SRO.SecOrDep.mp{2} /\ xs = dom RO.Or.mp{1} /\
   inp{1} = (elem{2}, SRO.SecOrDep.secret{2})).
auto.
wp.
if => //.
auto; smt.
qed.

local lemma ROLim_LOr_RO_Or_Conv_SRO_SecOrDep_hash :
  equiv
  [ROLim.LOr(RO.Or).hash ~ Conv(SRO.SecOrDep).hash :
   ={inp} /\
   G1.sec{1} = SRO.SecOrDep.secret{2} /\
   RO.Or.mp{1} = SRO.SecOrDep.mp{2} /\
   ROLim.LOr.inps{1} = SRO.SecOrDep.inps{2} /\
   ROLim.LOr.ctr{1} = SRO.SecOrDep.ctr{2} /\
   ROLim.LOr.inps{1} <= dom RO.Or.mp{1} ==>
   ={res} /\
   RO.Or.mp{1} = SRO.SecOrDep.mp{2} /\
   ROLim.LOr.inps{1} = SRO.SecOrDep.inps{2} /\
   ROLim.LOr.ctr{1} = SRO.SecOrDep.ctr{2} /\
   ROLim.LOr.inps{1} <= dom RO.Or.mp{1}].
proof.
proc.
inline SRO.SecOrDep.lhash.
seq 0 1 :
  (={inp} /\ inp0{2} = inp{2} /\
   G1.sec{1} = SRO.SecOrDep.secret{2} /\
   RO.Or.mp{1} = SRO.SecOrDep.mp{2} /\
   ROLim.LOr.inps{1} = SRO.SecOrDep.inps{2} /\
   ROLim.LOr.ctr{1} = SRO.SecOrDep.ctr{2} /\
   ROLim.LOr.inps{1} <= dom RO.Or.mp{1}).
auto.
if => //.
inline RO.Or.hash.
rcondf {1} 2.
auto; progress; smt.
auto.
if => //.
inline RO.Or.hash.
sp 3 2.
case (mem inp0{1} (dom RO.Or.mp{1})).
rcondf {1} 1; first auto.
rcondt {2} 1; first auto.
auto; smt.
rcondt {1} 1; first auto.
rcondf {2} 1; first auto.
wp; rnd; skip.
progress; smt.
auto.
qed.

local lemma G1_GDep_SecAdv_main :
  equiv
  [G1.main ~ SRO.GDep(SecAdv).main :
   true ==> ={res}].
proof.
proc.
inline ROLim.LOr(RO.Or).init RO.Or.init
  SRO.GDep(SecAdv).SA.main SRO.SecOrDep.init.
(* handle oracle and secret initialization *)
seq 4 6 :
  (G1.sec{1} = SRO.SecOrDep.secret{2} /\
   RO.Or.mp{1} = SRO.SecOrDep.mp{2} /\
   ROLim.LOr.ctr{1} = SRO.SecOrDep.ctr{2} /\
   ROLim.LOr.inps{1} = SRO.SecOrDep.inps{2} /\
   ROLim.LOr.inps{1} <= dom RO.Or.mp{1}).
swap {1} 4 -3.
auto; smt.
sim
  (: G1.sec{1} = SRO.SecOrDep.secret{2} /\
     RO.Or.mp{1} = SRO.SecOrDep.mp{2} /\
     ROLim.LOr.ctr{1} = SRO.SecOrDep.ctr{2} /\
     ROLim.LOr.inps{1} = SRO.SecOrDep.inps{2}) /
  (ROLim.LOr.inps{1} <= dom RO.Or.mp{1}) :
  (={b}).
exists* RO.Or.mp{1}; elim*; intros mp_L.
conseq* (G1_SRO_SecOrDep_hashElem_hash (dom mp_L)); smt.
conseq* (ROLim_LOr_RO_Or_Conv_SRO_SecOrDep_hash); smt.
qed.

local lemma G1_GDep_SecAdv &m :
  Pr[G1.main() @ &m : res] =
  Pr[SRO.GDep(SecAdv).main() @ &m : res].
proof.
byequiv G1_GDep_SecAdv_main; trivial.
qed.

(* G2 is like GIdeal, except that is uses Adv and inlining has been
   done *)

local module G2 = {
  var view : tp_view
  var tpHDB : hdb

  proc serverHashDB(db : db) : hdb = {
    var i : int;
    var elem : elem;
    var tag : tag;
    var hdb : tag list;
    db = Shuffle.shuffle(db);
    hdb = [];
    i = 0;
    while (i < length db) {
      elem = oget(nth db i);
      tag = Priv.Or.hash(elem);
      hdb = hdb ++ [tag];
      i = i + 1;
    }
    return hdb;
  }

  proc tpCountTag(tag : tag) : int = {
    var i, n : int;
    n = 0;
    i = 0;
    while (i < length tpHDB) {
      if (oget(nth tpHDB i) = tag) {
        n = n + 1;
      }
      i = i + 1;
    }
    view = view ++ [TPViewTagCount(tag, n)];
    return n;
  }

  proc clientLoop(qrys : qrys) : unit = {
    var i, n : int;
    var tag : tag;
    var qry : elem;
    i = 0;
    while (i < length qrys) {
      qry = oget(nth qrys i);
      tag = Priv.Or.hash(qry);
      n = tpCountTag(tag);
      i = i + 1;
    }
  }

  module A = Adv(ROLim.LOr(RO.Or))

  proc main() : bool = {
    var db : db;
    var qrys : qrys;
    var b : bool;
    var hdb : hdb;
    ROLim.LOr(RO.Or).init();
    Priv.Or.init();
    (db, qrys) = A.pre();
    tpHDB = serverHashDB(db);
    view = [TPViewHDB tpHDB];
    clientLoop(qrys);
    b = A.post(view);
    return b;
  }
}.

(* preparation for showing equivalence between SRO.GIndep(SecAdv) and
   G2 *)

local lemma Conv_SRO_SecOrIndep_ROLim_LOr_RO_Or_hash :
  equiv
  [Conv(SRO.SecOrIndep).hash ~ ROLim.LOr(RO.Or).hash :
   ={inp} /\
   SRO.SecOrIndep.lhmp{1} = RO.Or.mp{2} /\
   SRO.SecOrIndep.inps{1} = ROLim.LOr.inps{2} /\
   SRO.SecOrIndep.ctr{1} = ROLim.LOr.ctr{2} /\
   ROLim.LOr.inps{2} = dom RO.Or.mp{2} ==>
   ={res} /\
   SRO.SecOrIndep.lhmp{1} = RO.Or.mp{2} /\
   SRO.SecOrIndep.inps{1} = ROLim.LOr.inps{2} /\
   SRO.SecOrIndep.ctr{1} = ROLim.LOr.ctr{2} /\
   ROLim.LOr.inps{2} = dom RO.Or.mp{2}].
proof.
proc.
inline SRO.SecOrIndep.lhash.
seq 1 0 :
  (={inp} /\ inp0{1} = inp{1} /\
   SRO.SecOrIndep.lhmp{1} = RO.Or.mp{2} /\
   SRO.SecOrIndep.inps{1} = ROLim.LOr.inps{2} /\
   SRO.SecOrIndep.ctr{1} = ROLim.LOr.ctr{2} /\
   ROLim.LOr.inps{2} = dom RO.Or.mp{2}).
auto.
if => //.
inline RO.Or.hash.
rcondf {2} 2.
intros &m.
auto; smt.
auto; smt.
if => //.
inline RO.Or.hash.
rcondt {2} 4; first auto.
auto; smt.
auto.
qed.

local lemma GIndep_SecAdv_G2_main :
  equiv
  [SRO.GIndep(SecAdv).main ~ G2.main :
   true ==> ={res}].
proof.
proc.
inline SRO.GIndep(SecAdv).SA.main SRO.SecOrIndep.init
  ROLim.LOr(RO.Or).init RO.Or.init Priv.Or.init.
seq 5 4 :
  (SRO.SecOrIndep.lhmp{1} = RO.Or.mp{2} /\
   SRO.SecOrIndep.hmp{1} = Priv.Or.mp{2} /\
   SRO.SecOrIndep.inps{1} = ROLim.LOr.inps{2} /\
   SRO.SecOrIndep.ctr{1} = ROLim.LOr.ctr{2} /\
   ROLim.LOr.inps{2} = dom RO.Or.mp{2}).
auto; smt.
sim
  (: SRO.SecOrIndep.lhmp{1} = RO.Or.mp{2} /\
     SRO.SecOrIndep.hmp{1} = Priv.Or.mp{2} /\
     SRO.SecOrIndep.inps{1} = ROLim.LOr.inps{2} /\
     SRO.SecOrIndep.ctr{1} = ROLim.LOr.ctr{2}) /
  (ROLim.LOr.inps{2} = dom RO.Or.mp{2}) :
  (={b}).
conseq* (Conv_SRO_SecOrIndep_ROLim_LOr_RO_Or_hash); smt.
qed.

local lemma GIndep_SecAdv_G2 &m :
  Pr[SRO.GIndep(SecAdv).main() @ &m : res] =
  Pr[G2.main() @ &m : res].
proof.
byequiv GIndep_SecAdv_G2_main; trivial.
qed.

(* now we connect G1 and G2 by applying lemma SRO.GDepGIndep *)

local lemma G1_G2 &m :
  (forall (Or <: RO.OR{Adv}),
   islossless Or.hash => islossless Adv(Or).pre) =>
  (forall (Or <: RO.OR{Adv}),
   islossless Or.hash => islossless Adv(Or).post) =>
  `| Pr[G1.main() @ &m : res] -
     Pr[G2.main() @ &m : res] | <=
  limit%r / (2 ^ secLen)%r.
proof.
intros LossPre LossPost.
rewrite (G1_GDep_SecAdv &m).
rewrite - (GIndep_SecAdv_G2 &m).
apply (SRO.GDep_GIndep SecAdv &m).
apply SecAdv_Lossless_main.
apply LossPre.
apply LossPost.
qed.

local lemma GReal_G2 &m :
  (forall (Or <: RO.OR{Adv}),
   islossless Or.hash => islossless Adv(Or).pre) =>
  (forall (Or <: RO.OR{Adv}),
   islossless Or.hash => islossless Adv(Or).post) =>
  `| Pr[GReal(Adv).main() @ &m : res] -
     Pr[G2.main() @ &m : res] | <=
  limit%r / (2 ^ secLen)%r.
proof.
rewrite (GReal_G1 &m).
apply (G1_G2 &m).
qed.

local lemma G2_GIdeal_main :
  equiv
  [G2.main ~ GIdeal(Adv).main :
   true ==> ={res}].
proof.
proc.
inline GIdeal(Adv).tpAcceptHDB GIdeal(Adv).tpGetView.
(* handle initialization of oracles *)
seq 2 3 :
  (={glob ROLim.LOr(RO.Or), glob Priv.Or} /\
   GIdeal.tpView{2} = []).
inline *; auto.
(* handle call to pre *)
seq 1 1 :
  (={db, qrys, glob ROLim.LOr(RO.Or), glob Priv.Or, glob Adv} /\
   GIdeal.tpView{2} = []).
call
  (_ :
   ={glob ROLim.LOr(RO.Or), glob Priv.Or} ==>
   ={res, glob ROLim.LOr(RO.Or), glob Priv.Or, glob Adv}).
sim.
auto.
(* handle call to serverHashDB and setting of tpHDB and view *)
seq 2 4 :
  (={db, qrys, glob ROLim.LOr(RO.Or), glob Priv.Or, glob Adv} /\
   G2.tpHDB{1} = GIdeal.tpHDB{2} /\ G2.view{1} = GIdeal.tpView{2}).
wp.
call (_ : ={db, glob Priv.Or} ==> ={res, glob Priv.Or}).
sim.
auto.
(* handle call to clientLoop *)
seq 1 2 :
  (={db, qrys, glob ROLim.LOr(RO.Or), glob Priv.Or, glob Adv} /\
   G2.tpHDB{1} = GIdeal.tpHDB{2} /\ G2.view{1} = view{2});
  first sim.
(* call to post *)
sim.
qed.

local lemma G2_GIdeal &m :
  Pr[G2.main() @ &m : res] = Pr[GIdeal(Adv).main() @ &m : res].
proof.
byequiv G2_GIdeal_main => //.
qed.

lemma GReal_GIdeal' &m :
  (forall (Or <: RO.OR{Adv}),
   islossless Or.hash => islossless Adv(Or).pre) =>
  (forall (Or <: RO.OR{Adv}),
   islossless Or.hash => islossless Adv(Or).post) =>
  `| Pr[GReal(Adv).main() @ &m : res] -
     Pr[GIdeal(Adv).main() @ &m : res] | <=
  limit%r / (2 ^ secLen)%r.
proof.
rewrite - (G2_GIdeal &m).
apply (GReal_G2 &m).
qed.

end section.

lemma GReal_GIdeal
      (Adv <: ADV{ROLim.LOr(RO.Or), Priv.Or, GReal, GIdeal})
      &m :
  (forall (O <: RO.OR{Adv}),
   islossless O.hash => islossless Adv(O).pre) =>
  (forall (O <: RO.OR{Adv}),
   islossless O.hash => islossless Adv(O).post) =>
  `|Pr[GReal(Adv).main() @ &m : res] -
    Pr[GIdeal(Adv).main() @ &m : res]| <=
  limit%r / (2 ^ secLen)%r.
proof.
apply (GReal_GIdeal' Adv &m).
qed.
